{************************************************************************
*									*
*  J U M P -- JUMP.PAS							*
*									*
*  JUMP allows a user to login exactly as another user without a	*
*  password. It also allows a more restricted (non-exact) impersonation *
*  of another user.							*
*									*
*  Use of JUMP is restricted to specific categories of users: Systems	*
*  Programmers, Operators and any specifically authorised users.	*
*									*
*  The syntax of the JUMP command generally is				*
*									*
*      $ JUMP [username] [qualifiers]					*
*									*
*  For full documentation on JUMP, see the HELP file (JUMP.HLP), the	*
*  example access file (JUMP_ACCESS.DAT) and other files in the 	*
*  distribution kit.							*
*-----------------------------------------------------------------------*
*  ****  CAUTION: KERNEL-mode code fiddles things !!! ****		*
*									*
*  JUMP requires the following privileges:				*
*	CMEXEC, CMKRNL, DETACH (or IMPERSONATE), SYSNAM, SYSPRV 	*
*									*
*  INSTALL JUMP with these privileges if non-privileged users will be	*
*  running JUMP.							*
*									*
*  NOTE: The definition of privilege sets has been adopted for ease of	*
*  coding.  The source for the information is SYS$LIBRARY:STARLET.PAS.	*
*  The definition will need reviewing with each release of OpenVMS!	*
*-----------------------------------------------------------------------*
*									*
*  ***	CAUTION: JUMP has dependencies on the underlying architecture	*
*  ***		 (VAX, Alpha, IA64, X86) and the version of OpenVMS.	*
*  ***		 Any changes to either of these *REQUIRES* JUMP to be	*
*  ***		 re-linked.						*
*									*
*************************************************************************
*  Author and Maintainer: Jonathan Ridler.				*
*									*
*  This software is owned and maintained privately by the author,	*
*  Jonathan Ridler. It is NOT in any way produced, owned, supported,	*
*  maintained or endorsed by Hewlett Packard Pty Ltd or any of its	*
*  legal entities.							*
*									*
*  Email: vmsjump@internode.on.net					*
*************************************************************************
*************************************************************************
***								      ***
***			    LICENCE NOTICE			      ***
***								      ***
***  This software is COPYRIGHT (c) 1993-2022 Jonathan Ridler.	      ***
***  ALL RIGHTS RESERVED.					      ***
***								      ***
***  Please READ the file JUMP_LICENCE.TXT which contains the	      ***
***  complete Licence and all conditions of use for JUMP.	      ***
***								      ***
***  NOTE: By using JUMP or any part of the JUMP distribution for     ***
***	   any purpose, you are agreeing to all conditions of the     ***
***	   JUMP licence.					      ***
***								      ***
*************************************************************************
*************************************************************************
*									*
*  Revision History: See file CHANGES.TXT in the JUMP distribution.	*
*									*
************************************************************************}

[INHERIT ('SYS$LIBRARY:STARLET',
	  'SYS$LIBRARY:PASCAL$CLI_ROUTINES',
	  'SYS$LIBRARY:PASCAL$LIB_ROUTINES',
	  'SYS$LIBRARY:PASCAL$LBR_ROUTINES',
	  'SYS$LIBRARY:PASCAL$MAIL_ROUTINES',
	  'SYS$LIBRARY:PASCAL$STR_ROUTINES')]

PROGRAM Jump (OUTPUT, Logfile, Dbgfile) ;

CONST

Null = CHR (00) ;		{ NUL }
Bell = CHR (07) ;		{ BEL }
Lf   = CHR (10) ;		{ Linefeed }
Cr   = CHR (13) ;		{ Carriage Return }
Xon  = CHR (17) ;		{ XON }
Xoff = CHR (19) ;		{ XOFF }
Esc  = CHR (27) ;		{ ESC }

Io_Buflen = 1 * 512 ;		{ *WORDS* = 1 pair of 512 byte R/W blocks }
Rsts = 1 ;			{ Index to read status in pseudo-terminal buf }
Rcnt = 2 ;			{ Count of chars read in pseudo-terminal buf }
Rbuf = 3 ;			{ Actual read data area begins }
Wsts = Io_Buflen DIV 2 + 1 ;	{ Start of write buffer area and write status }
Wcnt = Wsts + 1 ;		{ Count of chars writ in pseudo-terminal buf }
Wbuf = Wsts + 2 ;		{ Actual read data area begins }
Ft_Buflen = Io_Buflen - 4 ;	{ *BYTES* Size of buffer for FTA device }
Py_Buflen = Io_Buflen ; 	{ *BYTES* Size of buffer for PYA device }

{ The effective length of a username is 12 characters. However, the internal
  fields are 32 characters.  Use 12 characters as the maximum length.

  ENSURE the constant is identical in the JUMP, JUMP_POKER and JUMP_PERSONA
  modules!  }

Max_Username_Len = 12 ;

Rightsize = 500 ;		{ Number of rightslist entries to retrieve }

Digits	   = ['0'..'9'] ;
Alphanum   = ['0'..'9','A'..'Z'] ;
Wildcard   = ['*','%'] ;
Symbol	   = Alphanum + ['$','_'] ;
Uic_Left   = ['[','<'] ;
Uic_Right  = [']','>'] ;
Separators = [',','-'] ;
Esc_Chars  = ['D','F','G','I','K','L','N','P','U','V'] ;

Max_Scm_Num = 1024 ;	{ Maximum number of commands for Single Command Mode }

TYPE

$BOOL  = [BIT(1),UNSAFE] BOOLEAN ;
$UBYTE = [BYTE] 0..255 ;
$UWORD = [WORD] 0..65535 ;
$UQUAD = [QUAD,UNSAFE] RECORD  L0 , L1 : UNSIGNED ;  END ;

Word_Ptr     = ^$UWORD ;
Unsigned_Ptr = [VOLATILE] ^UNSIGNED ;

Privilege = (Cmkrnl,	{  0: May change Mode to Kernel }
	     Cmexec,	{  1: May change Mode to Exec; MUST follow Cmkrnl }
	     Sysnam,	{  2: May insert in system logical name table }
	     Grpnam,	{  3: May insert in group l.n.tab; MUST follow Sysnam }
	     Allspool,	{  4: May allocate spooled device }
	     Detach,	{  5: May create detached processes (aka Impersonate) }
	     Diagnose,	{  6: May diagnose devices }
	     Log_Io,	{  7: May do logical I/O }
	     Group,	{  8: May affect other processes in same UIC group }
	     Acnt,	{  9: May suppress accounting }
	     Prmceb,	{ 10: May create permanent common event clusters }
	     Prmmbx,	{ 11: May create permanent mail box }
	     Pswapm,	{ 12: May change process mode }
	     Altpri,	{ 13: May set any priority value }
	     Setprv,	{ 14: May set any privilege bits }
	     Tmpmbx,	{ 15: May create temporary mailbox }
	     World,	{ 16: May affect other processes in the world }
	     Mount,	{ 17: May execute mount ACP functions }
	     Oper,	{ 18: Operator privilege }
	     Exquota,	{ 19: May exceed disk quotas }
	     Netmbx,	{ 20: May create network device }
	     Volpro,	{ 21: May override volume protection }
	     Phy_Io,	{ 22: May do physical I/O }
	     Bugchk,	{ 23: May make bug check error log entries }
	     Prmgbl,	{ 24: May create permanent global sections }
	     Sysgbl,	{ 25: May create system-wide global sections }
	     Pfnmap,	{ 26: May map to section by PFN }
	     Shmem,	{ 27: May allocate structures in shared memory }
	     Sysprv,	{ 28: Eligible for system protection field }
	     Bypass,	{ 29: May bypass UIC-based protection }
	     Syslck,	{ 30: May create system-wide locks }
	     Share,	{ 31: May assign channel to non-shared device }
	     Upgrade,	{ 32: May upgrade object integrity }
	     Downgrade, { 33: May downgrade object secrecy }
	     Grpprv,	{ 34: Group access via system protection field }
	     Readall,	{ 35: May read any object }
	     Import,	{ 36: May set classification for unlabeled object }
	     Audit,	{ 37: May direct audit to system security audit log }
	     Security,	{ 38: May perform security functions }
	     Pbit39, Pbit40, Pbit41, Pbit42, Pbit43, Pbit44, Pbit45,
	     Pbit46, Pbit47, Pbit48, Pbit49, Pbit50, Pbit51, Pbit52,
	     Pbit53, Pbit54, Pbit55, Pbit56, Pbit57, Pbit58, Pbit59,
	     Pbit60, Pbit61, Pbit62, Pbit63) ;	{ 39-63: dummy bits: 2 LONGs }

Privset = PACKED SET OF Privilege ;		{ Allow easy bit union, etc. }

{ Oper_Class defines the bits in the operator class mask. This should be
  reviewed with each major release of OpenVMS. }

Oper_Class = (Central,		{  0 }
	      Printer,		{  1 }
	      Tapes,		{  2 }
	      Disks,		{  3 }
	      Devices,		{  4 }
	      Cards,		{  5 }
	      Network,		{  6 }
	      Cluster,		{  7 }
	      Opc_Security,	{  8 = SECURITY ... avoid same name as priv! }
	      Reply,		{  9 }
	      Software, 	{ 10 }
	      License,		{ 11 }
	      Oper1,		{ 12 = USER1 }
	      Oper2,		{ 13 = USER2 }
	      Oper3,		{ 14 = USER3 }
	      Oper4,		{ 15 = USER4 }
	      Oper5,		{ 16 = USER5 }
	      Oper6,		{ 17 = USER6 }
	      Oper7,		{ 18 = USER7 }
	      Oper8,		{ 19 = USER8 }
	      Oper9,		{ 20 = USER9 }
	      Oper10,		{ 21 = USER10 }
	      Oper11,		{ 22 = USER11 }
	      Oper12) ; 	{ 23 = USER12 }

Access_Status = (Granted,Denied,Unspecified) ;	{ Status from access file }

Architecture_Type = (Vax,Alpha,Integrity,X86) ;	{ Past, present ... and future!}

Status_Block_Type = [UNSAFE] PACKED ARRAY [1..4] OF $UWORD ;

Item_List_Cell = RECORD
		   CASE INTEGER OF
		     1: (			{ Normal Cell }
			 Buffer_Length : $UWORD ;
			 Item_Code     : $UWORD ;
			 Buffer_Addr   : UNSIGNED ;
			 Return_Addr   : UNSIGNED
			) ;
		     2: (			{ Terminator }
			 Terminator    : UNSIGNED
			) ;
		 END ;

Item_List_Template (Count:INTEGER) = ARRAY [1..Count] OF Item_List_Cell ;

Username_Type = PACKED ARRAY [1..Max_Username_Len] OF CHAR ;

Just_A_String = VARYING [1024] OF CHAR VALUE ZERO ;

Rights_Array = ARRAY [1..Rightsize] OF $UQUAD ;

{ Signal and Mechanism parameters ... }

Sig_Args  = ARRAY [0..100] OF INTEGER ;
Mech_Args = ARRAY [0..(SIZE(CHF2$TYPE)-4)DIV 4] OF [UNSAFE] INTEGER ;

Io_Buffer = [ALIGNED(9),STATIC,UNSAFE]			{ Pagelet aligned }
		RECORD
		  CASE BOOLEAN OF
		  TRUE:  (Raw : ARRAY [1..Io_Buflen]   OF $UWORD  VALUE ZERO) ;
		  FALSE: (One : ARRAY [1..Io_Buflen*2] OF CHAR	  VALUE ZERO) ;
		END ;

Terminal_Chars = PACKED RECORD
		   Tt_Class    : [POS(0)]  $UBYTE ;
		   Tt_Type     : [POS(8)]  $UBYTE ;
		   Tt_Width    : [POS(16)] $UWORD ;
		   Tt_Devchar  : [POS(32)] TT$TYPE ;
		   Tt_Devchar2 : [POS(64)] TT2$TYPE ;
		 END ;

Prtctl_Type = PACKED RECORD
		TTY$V_PC_NOTIME       : [POS(0)]  $BOOL ;
		TTY$V_PC_DMAENA       : [POS(1)]  $BOOL ;
		TTY$V_PC_DMAAVL       : [POS(2)]  $BOOL ;
		TTY$V_PC_PRMMAP       : [POS(3)]  $BOOL ;
		TTY$V_PC_MAPAVL       : [POS(4)]  $BOOL ;
		TTY$V_PC_XOFAVL       : [POS(5)]  $BOOL ;
		TTY$V_PC_XOFENA       : [POS(6)]  $BOOL ;
		TTY$V_PC_NOCRLF       : [POS(7)]  $BOOL ;
		TTY$V_PC_BREAK	      : [POS(8)]  $BOOL ;
		TTY$V_PC_PORTFDT      : [POS(9)]  $BOOL ;
		TTY$V_PC_NOMODEM      : [POS(10)] $BOOL ;
		TTY$V_PC_NODISCONNECT : [POS(11)] $BOOL ;
		TTY$V_PC_SMART_READ   : [POS(12)] $BOOL ;
		TTY$V_PC_ACCPORNAM    : [POS(13)] $BOOL ;
		TTY$V_PC_MULTISESSION : [POS(15)] $BOOL ;
	      END ;

Notify_Mask = RECORD
		CASE INTEGER OF
		1: (All_Bits	: UNSIGNED ;) ; 	{ 32 bits - roomy! :) }
		2: (After	: [BIT, POS(0)] BOOLEAN ;    { EXACT completed }
		    Before	: [BIT, POS(1)] BOOLEAN ;    { EXACT initiated }
		    Include_Log : [BIT, POS(2)] BOOLEAN ;    { Log in MAIL }
		    By_Mail	: [BIT, POS(3)] BOOLEAN ;    { Send MAIL }
		    By_Opcom	: [BIT, POS(4)] BOOLEAN ;    { OPCOM message }
		    Vamoose	: [BIT, POS(5)] BOOLEAN ;) ; { Exit if problem }
	      END ;

Debug_Mask = RECORD
	       CASE INTEGER OF
	       1: (All_Bits : UNSIGNED ;) ;    { 32 bits - roomy! :) }
	       2: (Info     : [BIT, POS(0)]  BOOLEAN ;	   { General Info }
		   D1	    : [BIT, POS(1)]  BOOLEAN ;	   { Diagnostic info }
		   D2	    : [BIT, POS(2)]  BOOLEAN ;	   { Diagnostic info }
		   D3	    : [BIT, POS(3)]  BOOLEAN ;	   { Diagnostic info }
		   D4	    : [BIT, POS(4)]  BOOLEAN ;	   { Diagnostic info }
		   D5	    : [BIT, POS(5)]  BOOLEAN ;	   { Diagnostic info }
		   D6	    : [BIT, POS(6)]  BOOLEAN ;	   { Diagnostic info }
		   D7	    : [BIT, POS(7)]  BOOLEAN ;	   { Diagnostic info }
		   D8	    : [BIT, POS(8)]  BOOLEAN ;	   { Diagnostic info }
		   D9	    : [BIT, POS(9)]  BOOLEAN ;	   { Diagnostic info }
		   X1	    : [BIT, POS(11)] BOOLEAN ;	   { Change execution }
		   X2	    : [BIT, POS(12)] BOOLEAN ;	   { Change execution }
		   X3	    : [BIT, POS(13)] BOOLEAN ;	   { Change execution }
		   X4	    : [BIT, POS(14)] BOOLEAN ;	   { Change execution }
		   X5	    : [BIT, POS(15)] BOOLEAN ;	   { Change execution }
		   X6	    : [BIT, POS(16)] BOOLEAN ;	   { Change execution }
		   X7	    : [BIT, POS(17)] BOOLEAN ;	   { Change execution }
		   X8	    : [BIT, POS(18)] BOOLEAN ;	   { Change execution }
		   X9	    : [BIT, POS(19)] BOOLEAN ;) ;  { Change execution }
	     END ;

CONST

Required_Privs = [Cmexec,Cmkrnl,Detach,Sysnam,Sysprv] ; { Needed to run JUMP }
Single_Cmd_Mode_Privs = [Group,World] ; 		{ Extra privs for SCM }

{************************************************************************
 ****	   Update the Production_Version status as appropriate	     ****
 ************************************************************************}

Production_Version = TRUE ;	{ Set TRUE only when ready for PRODUCTION use }

VAR

{************************************************************************
 ****  INCREMENT the BUILD LEVEL every time the module is changed!!  ****
 ************************************************************************}

Main_Bl : [GLOBAL] VARYING [12] OF CHAR := '1.055' ;

{ Other module build levels }

Fiddler_Bl ,			{ Build level for Fiddler module }
Persona_Bl ,			{ Build level for Persona modules }
Poker_Bl ,			{ Build level for Poker module }
Vax_Bl				{ Build level for VAX modules }
: [EXTERNAL] VARYING [12] OF CHAR ;

{************************************************************************
 **** Keep the version up-to-date and in sync with the JUMP.OPT file ****
 ************************************************************************}

Jump_Version
: VARYING [255] OF CHAR := 'JUMP V7.0 2022-11-25 (25-Nov-2022)' ;

Sysprog ,			{ SETPRV priv or group <= MAXSYSGROUP ? }
Operator ,			{ OPER priv ? }
Log ,				{ Log success messages? }
Debug , 			{ Debug messages and execution control? }
Allow_Debug ,			{ Allow this non-privileged user to debug? }
User_Debug_Opcom ,		{ Send message to OPCOM when Allow_Debug on? }
Alter_Ego ,			{ Change Username? }
Transmute ,			{ Change UIC, etc.? }
Auditing ,			{ Audit jumps? }
Real_Mccoy ,			{ Use a pseudo-terminal and *really* do it!? }
Double_Check ,                  { Double check general user's access? }
Narcissus ,			{ Allow users to jump to themselves? }
Broken_Chain ,			{ Attempt to chain when chaining prohibited? }
Figment ,			{ Allow username NOT in UAF? }
Stamp_Uaf ,			{ Update UAF Last Interactive Login + Fails? }
Suspect ,			{ Trying to subvert Secure Mode? }
Houdini ,			{ Allow override of Secure Mode? }
High_Security , 		{ Secure file open? }
Secure_Mode ,			{ Secure mode requested? }
System_Secure_Mode ,		{ Secure mode mandated? }
Secure_Logical			{ Logical name is defined /SYSTEM /EXEC ? }
: BOOLEAN := FALSE ;

Chain ,                         { JUMP to target user's allowed JUMP targets? }
Secure_User_Dir 		{ JUMP_USER_DIR defined by System Manager? }
: BOOLEAN := TRUE ;

Viewable ,			{ Text entered in terminal buffer visible? }
Visible 			{ Text entered in terminal buffer visible? }
: [VOLATILE] BOOLEAN := TRUE ;

Use_Persona ,			{ Use the PERSONA system calls, if available? }
Single_Cmd_Mode ,		{ Single Command Mode for EXACT }
Open_New_File , 		{ Open new file in Special File Open? }
Debug_To_Terminal ,		{ Debug output going to a terminal device? }
Record_Session ,		{ Make a recording of an EXACT jump? }
Escape_Hatch ,			{ Escape character defined for EXACT jump? }
Jeronimo			{ Have we pulled the ripcord and escaped? }
: [VOLATILE] BOOLEAN := FALSE ;

Arb_Full_Support ,		{ Access Rights Block: ARB_SUPPORT = 3 ? }
Psb_Available			{ OpenVMS V7.2+ ? ==> Persona Security Block! }
: [GLOBAL,VOLATILE] BOOLEAN := FALSE ;

Ripcord : [VOLATILE] CHAR := 'D' ;	{ Escape character for EXACT jump }

Default_Scm ,				{ Default SCM command number }
Scm_Count ,				{ Number of commands in 'U' option }
Scm_Selection , 			{ Command number selected for SCM }
Max_Sys_Group : INTEGER := 0 ;	{ Maximum UIC group with system privileges }

Oper_Class_Mask : UNSIGNED := 0 ;

Mail_Error_Status : [VOLATILE] UNSIGNED := 0 ;	{ MAIL condition value }

Appcrash : REAL := 0 ;			{ Just a real to allow JUMP to crash }

Acme_Agent_Name ,    { External authentication Persona extension name }
Orig_User ,                                             { Invoker's username }
New_User : [GLOBAL,VOLATILE] VARYING [Max_Username_Len] OF CHAR
			:= PAD ('',' ',Max_Username_Len) ; { Target user }

Time_Str : VARYING [23] OF CHAR := '' ; { Time string for filename creation }

Jump_Modules ,						{ Module versions }
Open_File_Name ,					{ File spec }
Command : VARYING [255] OF CHAR := '' ; 		{ Input command line }

Orig_Uic ,						{ UIC of caller }
New_Uic : [GLOBAL,VOLATILE] UIC$TYPE := ZERO ;		{ UIC of target user }

Terminal ,						{ Audit this }
Port ,							{ Audit this }
Physical_Device ,					{ Audit this }
Def_Dev : [VOLATILE] VARYING [64] OF CHAR := '' ;	{ Default device }

Target_Attributes ,				{ Target process attributes }
Help_Library ,					{ Help Library filespec }
Help_Topic ,					{ Optional topic for help }
Secure_Directory ,				{ Directory for secure logs }
User_Directory ,				{ Directory for insecure logs }
Session_Log ,					{ Filespec of session log }
Access_List ,					{ Filespec }
Audit_Trail : VARYING [255] OF CHAR := '' ;	{ Filespec }

Single_Cmd ,					{ Single Command for EXACT }
Dbgfile_Name ,					   { Debug output filename }
Def_Dir : [VOLATILE] VARYING [255] OF CHAR := '' ;     { Default directory }

Notify_Maillist : VARYING [1022] OF CHAR := '' ;	{ Who to notify }

Valid_Scm_Numbers
: ARRAY [1..Max_Scm_Num] OF BOOLEAN := ZERO ;		{ Allowed SCM nums }

Vms_Arch ,						{ VMS architecture }
Vms_Version ,						{ Version of VMS }
Vms_Version_Num 					{ Just the number bit }
: VARYING [15] OF CHAR := ZERO ;

Uic_Str : [GLOBAL] VARYING [15] OF CHAR := ZERO ;	{ String form }

Eq_Id_Str ,						{ UIC string handling }
Id_Str	: [GLOBAL] VARYING [32] OF CHAR := ZERO ;	{ UIC string handling }

Minor_Privs ,						{ Not major privs! :) }
Def_Priv ,                                              { Target's def  privs }
Auth_Priv : [UNSAFE,VOLATILE]  Privset := ZERO ;        { Target's auth privs }

Oper_Classes : PACKED SET OF Oper_Class := ZERO ;

Vmsauth_Flag ,
Extauth_Flag : [EXTERNAL] BOOLEAN ;

Caller_Flags ,
Target_Flags : [VOLATILE]  FLAGS$TYPE := ZERO ;         { Target's UAF flags }

Notify : Notify_Mask := ZERO ;			{ Notification flags }
Bugger : [GLOBAL,VOLATILE] Debug_Mask := ZERO ; { Debug flags }

Architecture : Architecture_Type := Alpha ;	{ Assume Alpha }

Jib_User_Ptr : [VOLATILE] ^[VOLATILE] Username_Type := NIL ;

Jib_Ptr ,
Jib_Uic_Ptr ,
Uic_Ptr : Unsigned_Ptr := NIL ;

Pchan_Created , 				{ Has Pchan been created? }
Pseudo_Ft : [VOLATILE] BOOLEAN := FALSE ;	{ Is pseudo-terminal FTA0:? }

Pchan , 					{ Pseudo-terminal channel }
Rchan , 					{ Real terminal channel }
Mchan : [VOLATILE] $UWORD := 0 ;		{ Mailbox channel }

Pdev : VARYING [12] OF CHAR := '' ;		{ Pseudo-terminal device name }

Pbuf_Range : ARRAY [1..2] OF UNSIGNED := ZERO ; { Quasi descriptor of I/O buf }

Mbbuf : ARRAY [1..ACC$K_TERMLEN] OF $UBYTE := ZERO ;	{ Termination MBX buf }

Piosb : [VOLATILE] Status_Block_Type := ZERO ;	{ IOSB for pseudo-terminal IOs }

Buffer : [VOLATILE] Io_Buffer := ZERO ; 	{ Pseudo-terminal I/O buffer }

Rchars : Terminal_Chars := ZERO ;		{ Device chars of real term }

Exit_Rst ,					{ Exit Handler status }
Pid ,						{ Current process PID }
Master_Pid ,					{ Current process master PID }
Pseudo_Pid ,					{ Pseudo-terminal process PID }
Job_Proc_Cnt_Now ,				{ Job subproc count now }
Job_Proc_Cnt ,					{ Initial job subproc count }
Proc_Cnt_Now ,					{ Proc subproc count now }
Proc_Cnt : [VOLATILE] UNSIGNED := 0 ;		{ Current proc subproc count }

Proc_Cur_Priv ,                                       { Caller's current privs }
Proc_Perm_Priv ,                                      { Caller's default privs }
Proc_Auth_Priv : [UNSAFE,VOLATILE]  Privset := ZERO ; { Caller's auth'd  privs }

Caller_Rights ,                                       { Caller's proc rights }
Target_Rights ,                                       { Target's proc rights }
System_Rights : Rights_Array := ZERO ;		      { System rights }

Time_Now : TIMESTAMP ;				      { Real time now }

Login_Time : $UQUAD := ZERO ;                         { Caller's login time }

Login_Time_Str : VARYING [23] OF CHAR := '' ;         { Caller's login time }

Process_Name : VARYING [16] OF CHAR := '' ;           { Caller's process name }

Null_List : Item_List_Template(1) := ZERO ;	      { Empty item list }

Open_File_Fab : FAB$TYPE := ZERO ;		{ So we can see file info }

Dbgfile ,					{ Debug output file }
Logfile : [GLOBAL,VOLATILE] TEXT ;		{ Logfile for EXACT session }

{ The values "Jump__*" are condition codes used exclusively by JUMP.
  Details can be found in the Message file (JUMP_MSG.MSG). }

Jump__Abortx9 ,
Jump__Badaccfil ,
Jump__Badaudit ,
Jump__Baddata ,
Jump__Baddbgfil ,
Jump__Badinclude ,
Jump__Badlogfil ,
Jump__Badnotify ,
Jump__Badoperclass ,
Jump__Badprivset ,
Jump__Baduser ,
Jump__Conflict ,
Jump__Crashrequested ,
Jump__Debugfile ,
Jump__Debugon ,
Jump__Default ,
Jump__Denied ,
Jump__Devnotprd ,
Jump__Disabled ,
Jump__Fixnotify ,
Jump__Forcepss ,
Jump__Intabort1 ,
Jump__Invescchar ,
Jump__Invlnm ,
Jump__Invuser ,
Jump__Ivident ,
Jump__Jumped ,
Jump__Mailfail ,
Jump__Modules ,
Jump__Nochain ,
Jump__Noescape ,
Jump__Noinsub ,
Jump__Nolnmforscm ,
Jump__Noovrsnglcmd ,
Jump__Nopriv ,
Jump__Nopseudo ,
Jump__Nospawn ,
Jump__Nosub ,
Jump__Notusepss ,
Jump__Restrict ,
Jump__Return ,
Jump__Sameuic ,
Jump__Sameuser ,
Jump__Scmcmddef ,
Jump__Scmnotsel ,
Jump__Scmreqcmd ,
Jump__Selnotreq ,
Jump__Setuser ,
Jump__Transfer ,
Jump__Unkscmnum ,
Jump__Unsupported ,
Jump__Userabort ,
Jump__Version : [EXTERNAL,VALUE] UNSIGNED ;

{ Some system and compiler-related values not found in STARLET. }

CCB$L_UCB ,
CCB$L_CHAN ,
CCB$K_LENGTH ,
UCB$L_TL_PHYUCB ,
UCB$W_TT_PRTCTL ,
UCB$L_TT_ACCPORNAM ,
JIB$T_USERNAME ,
PAS$K_SUCCESS ,
PAS$K_FILNOTFOU : [EXTERNAL,VALUE] UNSIGNED ;

Iss__C_Arb_Full ,		{ = ISS$C_ARB_FULL }
Syi___Arb_Support ,		{ = SYI$_ARB_SUPPORT }
CTL$GL_PCB ,
CTL$GL_CCBBASE ,
CTL$GA_CCB_TABLE : [EXTERNAL] UNSIGNED ;

CTL$T_USERNAME : [EXTERNAL,VOLATILE] Username_Type ;


PROCEDURE Check_Extauth_Flags (Flags : FLAGS$TYPE) ; EXTERNAL ;


PROCEDURE  Set_Default_Dir (Def_Dir : VARYING [Len] OF CHAR); EXTERNAL ;


FUNCTION Condition_Code (Pascal_Status : UNSIGNED) : UNSIGNED ;

{ Convert a Pascal status to a condition code. }

  BEGIN 	{ Condition_Code }
  Condition_Code := 16#218644 + Pascal_Status * 8 ;
  END ; 	{ Condition_Code }


[ASYNCHRONOUS] FUNCTION General_Error_Handler (
			VAR Sa : Sig_Args;
			VAR Ma : Mech_Args) : [UNSAFE] INTEGER ;

{ A simple handler that allows conditions to be debugged more easily
  when traceback cannot be set. }

  BEGIN
  IF NOT ODD (Sa[1])  THEN
     WRITELN (Dbgfile,'*** Condition Value ',HEX(Sa[1],8),
		      ' has been signalled from within JUMP') ;
  General_Error_Handler := SS$_RESIGNAL ;	{ Let VMS handle the error }
  END ;


[GLOBAL,ASYNCHRONOUS] PROCEDURE Zip (Rst1 : UNSIGNED ;
				     Rst2 : UNSIGNED := SS$_NORMAL) ;

  { Just a shorthand for signalling return statuses. The second return
    status is optional. Normally, the second value will be IOSB[1]. }

  BEGIN 	{ Zip }
  IF NOT ODD (Rst1)  THEN
     LIB$SIGNAL (Rst1)
   ELSE
  IF NOT ODD (Rst2)  THEN
     LIB$SIGNAL (Rst2) ;
  END ; 	{ Zip }


FUNCTION Special_File_Open (VAR Fab : FAB$TYPE ;
			    VAR Rab : RAB$TYPE ;
			    VAR Log : TEXT) : UNSIGNED ;

  { Special file open. Hijack the NAM block to determine the actual filespec
    of the file. If high security, use executive mode logical names only. }

  VAR Rst : UNSIGNED := 0 ;
      Nam : ^NAM$TYPE := ZERO ;

  BEGIN 	{ Special_File_Open }
  IF High_Security OR Secure_Mode OR Secure_User_Dir  THEN
     Fab.FAB$V_LNM_MODE := PSL$C_EXEC ;       { Executive mode only! }

  High_Security := FALSE ;		      { Reset to default }
  Open_File_Fab := ZERO ;		      { In case the action fails }

  Nam := Fab.FAB$L_NAM::Pointer ;

  IF Open_New_File  THEN
     BEGIN
     Open_New_File := FALSE ;		      { Reset to default }
     Rst := $CREATE (Fab) ;
     END
   ELSE
     Rst := $OPEN (Fab) ;

  IF ODD (Rst)	THEN
     BEGIN
     Open_File_Fab := Fab ;
     STR$COPY_R (%DESCR Open_File_Name,Nam^.NAM$B_RSL,%IMMED Nam^.NAM$L_RSA) ;
     Rst := $CONNECT (Rab)
     END ;

  Special_File_Open := Rst ;
  END ; 	{ of Special_File_Open }


FUNCTION Get_Logical_Name (
		Lognam	: [CLASS_S] PACKED ARRAY [L1..U1:INTEGER] OF CHAR ;
		Default : VARYING [Sz1] OF CHAR ;
		VAR Actual : VARYING [Sz2] OF CHAR ;
		Table	: [CLASS_S] PACKED ARRAY [L2..U2:INTEGER] OF CHAR :=
				'LNM$SYSTEM' ;
		Mode	: $UBYTE := PSL$C_EXEC ;
		Lnm_Index  : UNSIGNED := 0
		) : UNSIGNED ;

{ Get the translation of the logical name specified.  If it does not exist,
  use the default value if the index is zero, or flag a non-existant index. }

  VAR Rst : UNSIGNED := 0 ;
      Attributes : LNM$TYPE := ZERO ;
      Actual_Mode : $UBYTE := PSL$C_USER ;
      Actual_Table : VARYING [32] OF CHAR := '' ;
      Item_List  : Item_List_Template (6) := ZERO ;

  BEGIN       { Get_Logical_Name }
  Item_List[1].Buffer_Length := 4 ;
  Item_List[1].Item_Code     := LNM$_INDEX ;
  Item_List[1].Buffer_Addr   := IADDRESS (Lnm_Index) ;
  Item_List[1].Return_Addr   := 0 ;

  Item_List[2].Buffer_Length := SIZE (Attributes) ;
  Item_List[2].Item_Code     := LNM$_ATTRIBUTES ;
  Item_List[2].Buffer_Addr   := IADDRESS (Attributes) ;
  Item_List[2].Return_Addr   := 0 ;

  Item_List[3].Buffer_Length := SIZE (Actual_Table.BODY) ;
  Item_List[3].Item_Code     := LNM$_TABLE ;
  Item_List[3].Buffer_Addr   := IADDRESS (Actual_Table.BODY) ;
  Item_List[3].Return_Addr   := IADDRESS (Actual_Table.LENGTH) ;

  Item_List[4].Buffer_Length := 1 ;
  Item_List[4].Item_Code     := LNM$_ACMODE ;
  Item_List[4].Buffer_Addr   := IADDRESS (Actual_Mode) ;
  Item_List[4].Return_Addr   := 0 ;

  Item_List[5].Buffer_Length := SIZE (Actual.BODY) ;
  Item_List[5].Item_Code     := LNM$_STRING ;
  Item_List[5].Buffer_Addr   := IADDRESS (Actual.BODY) ;
  Item_List[5].Return_Addr   := IADDRESS (Actual.LENGTH) ;

  Item_List[6].Terminator    := 0 ;   { Terminate the item list }

  Rst := $TRNLNM (Attr	 := %REF LNM$M_CASE_BLIND,
		  Tabnam := Table,
		  Lognam := Lognam,
		  Acmode := %REF Mode,
		  Itmlst := Item_List) ;
  Get_Logical_Name := Rst ;

  Secure_Logical := FALSE ;

  IF Rst = SS$_NOLOGNAM  THEN
     Actual := Default
   ELSE
  IF Rst = SS$_NORMAL  THEN
     BEGIN
     IF NOT Attributes.LNM$V_EXISTS  THEN
	Get_Logical_Name := SS$_VALNOTVALID ;  { Index not found - tell caller }
     IF (Actual_Mode = PSL$C_EXEC) AND (Actual_Table = 'LNM$SYSTEM')  THEN
	Secure_Logical := TRUE ;
     END
   ELSE
  IF Rst = SS$_BUFFEROVF  THEN		     { Do nothing - caller must handle }
   ELSE
  Zip (Rst) ;
  END ;       { of Get_Logical_Name }


PROCEDURE Define_Logical_Name (
		Lognam : [CLASS_S] PACKED ARRAY [L1..U1:INTEGER] OF CHAR ;
		Xlate  : [CLASS_S] PACKED ARRAY [L2..U2:INTEGER] OF CHAR ;
		Table  : [CLASS_S] PACKED ARRAY [L3..U3:INTEGER] OF CHAR :=
				'LNM$SYSTEM' ;
		Mode   : $UBYTE := PSL$C_EXEC) ;

{ Define the logical name specified.  BEWARE!  SYSNAM privilege is REQUIRED
  for supervisor mode or greater. }

  VAR Rst : UNSIGNED := 0 ;
      Item_List : Item_List_Template (2) := ZERO ;

  BEGIN       { Define_Logical_Name }
  Item_List[1].Buffer_Length := U2 ;
  Item_List[1].Item_Code     := LNM$_STRING ;
  Item_List[1].Buffer_Addr   := IADDRESS (Xlate) ;
  Item_List[1].Return_Addr   := 0 ;

  Item_List[2].Terminator    := 0 ;

  Rst := $CRELNM (Tabnam := Table,
		  Lognam := Lognam,
		  Acmode := Mode,
		  Itmlst := Item_List) ;
  Zip (Rst) ;
  END ;       { of Define_Logical_Name }


FUNCTION Oprmsg (Message : VARYING [Len] OF CHAR ;
		 Oper_Class_Mask : UNSIGNED := 16#00FFFFFF ;	{ 24 bits only }
		 Reply : BOOLEAN := FALSE) : BOOLEAN ;

{ Send an operator message (REQUEST).  Limit size to maximum (128 chars).
  If required, wait for a REPLY and flag if an ABORT was received. }

  VAR	Rst : UNSIGNED := 0 ;
	Mbx_Chan : $UWORD := 0 ;
	Msg_Text : VARYING [128] OF CHAR := ZERO ;
	Opr_Reply : OPC$TYPE := ZERO ;
	Iosb : Status_Block_Type := ZERO ;
	Msg_Dsc : DSC1$TYPE := ZERO ;
	Msg : PACKED RECORD
		Msg_Type : $UBYTE ;
		Msg_Targ : [BYTE(3)] 0..2**24-1 ;
		Msg_Rqst : UNSIGNED ;
		Mess : PACKED ARRAY [1..128] OF CHAR ;
	      END := ZERO ;

  BEGIN 	{ Oprmsg }
  Oprmsg := TRUE ;				{ Default }

  IF Message.LENGTH > 128  THEN
     Msg_Text := SUBSTR (Message.BODY,1,128)	{ Max msg size }
   ELSE
     Msg_Text := Message ;
  Msg.Msg_Type := OPC$_RQ_RQST ;
  Msg.Msg_Targ := Oper_Class_Mask ;
  Msg.Msg_Rqst := 0 ;
  Msg.Mess     := Msg_Text.BODY ;
  Msg_Dsc.DSC$W_MAXSTRLEN := 8 + Msg_Text.LENGTH ;
  Msg_Dsc.DSC$B_DTYPE	  := DSC$K_DTYPE_T ;	{ Not essential ... but neat! }
  Msg_Dsc.DSC$A_POINTER   := IADDRESS (Msg) ;
  Msg_Dsc.DSC$B_CLASS	  := DSC$K_CLASS_S ;

  IF Reply  THEN
     BEGIN
     Rst := $CREMBX (Chan := Mbx_Chan) ;		{ Create mailbox }
     Zip (Rst) ;
     END ;

  Rst := $SNDOPR (%REF Msg_Dsc,%IMMED Mbx_Chan) ;	{ Send to Operator }
  Zip (Rst) ;

  IF Reply  THEN
     BEGIN
     Rst := $QIOW (Chan := Mbx_Chan,			{ Read from mailbox }
		   Func := IO$_READVBLK,
		   P1	:= Opr_Reply,
		   P2	:= SIZE (Opr_Reply),
		   Iosb := Iosb) ;
     Zip (Rst,Iosb[1]) ;

     Rst := Opr_Reply.OPC$W_MS_STATUS + 65536 * OPCOM$_FACILITY ;
     IF Rst = OPC$_RQSTABORT  THEN
	Oprmsg := FALSE
      ELSE
     Zip (Rst) ;

     Rst := $DASSGN (Mbx_Chan) ;			{ Delete mailbox }
     Zip (Rst) ;
     END ;
  END ; 	{ of Oprmsg }


[GLOBAL,ASYNCHRONOUS] FUNCTION	Str_Compress (
			Source : [CLASS_S] PACKED ARRAY [L..U:INTEGER] OF CHAR ;
			VAR  Dest : VARYING [D] OF CHAR ;
			Collapse : BOOLEAN := FALSE) : UNSIGNED ;

{ Compress a string by removing leading and trailing white space (blanks and
  tabs), and replacing multiple consecutive white space with a single blank.
  If collapse is set, remove ALL white space. }

  CONST  Blanks = [' ',''(9),''(0)] ;
	 Maxsize = 1024 ;
	 Warn_Inpstrtru = UAND (LIB$_INPSTRTRU,%Xfffffff8) ;	{ Warning only }

  VAR  S , J , K : INTEGER := 0 ;
       Done : BOOLEAN := FALSE ;
       Spacer : VARYING [1] OF CHAR := ' ' ;
       Dstr, Sstr : VARYING [Maxsize] OF CHAR := '' ;

  BEGIN 	{ Str_Compress }
  Str_Compress := SS$_NORMAL ;		{ Presume so }
  S := LENGTH (Source) ;

  IF S = 0  THEN			{ Nothing passed }
     Dest := ''
   ELSE
     BEGIN
     IF S > Maxsize  THEN
	BEGIN
	Sstr := SUBSTR (Source,1,Maxsize) ;
	Str_Compress := INT(Warn_Inpstrtru) ;
	END
      ELSE
	Sstr := Source ;

     IF Collapse  THEN	Spacer := '' ;

     IF FIND_MEMBER (Source,Blanks) = 0  THEN		{ Nothing to change }
	Dstr := Sstr
      ELSE
	WHILE NOT Done	DO
	  BEGIN
	  J := FIND_NONMEMBER (Sstr,Blanks) ;
	  IF J = 0  THEN
	     Done := TRUE
	   ELSE
	     BEGIN
	     Sstr := SUBSTR (Sstr,J,Sstr.LENGTH-J+1) ;
	     K := FIND_MEMBER (Sstr,Blanks) ;
	     IF K = 0  THEN
		BEGIN
		Dstr := Dstr + Sstr ;
		Done := TRUE ;
		END
	      ELSE
		BEGIN
		Dstr := Dstr + SUBSTR (Sstr,1,K-1) + Spacer ;
		Sstr := SUBSTR (Sstr,K,Sstr.LENGTH-K+1) ;
		END ;
	     END ;
	  END ; 	{ of While }

     IF Dstr <> ''  THEN
	IF Dstr[Dstr.LENGTH] = ' '  THEN
	   Dstr:= SUBSTR (Dstr,1,Dstr.LENGTH-1) ;

     IF Dstr.LENGTH <= D  THEN
	Dest := Dstr
      ELSE
	BEGIN
	Dest := SUBSTR (Dstr,1,D) ;
	Str_Compress := LIB$_OUTSTRTRU ;
	END ;
     END ;
  END ; 	{ of Str_Compress }


PROCEDURE Give_Help ;

{ M'aidez!  Allow a subtopic specification. }

  VAR  Rst : UNSIGNED := 0 ;

  BEGIN 	{ Give_Help }
  Rst := LBR$OUTPUT_HELP (Output_Routine := %IMMED LIB$PUT_OUTPUT,
			  Line_Desc	 := 'JUMP ' + Help_Topic,
			  Library_Name	 := Help_Library,
			  Input_Routine  := %IMMED LIB$GET_INPUT) ;
  Zip (Rst) ;
  END ; 	{ Give_Help }


[GLOBAL,ASYNCHRONOUS] FUNCTION Str_Privs (Privs : Privset) : Just_A_String ;

{ Create a string representing the privileges in a privilege set. }

  VAR	P : Privilege ;
	T1 , T2 : Just_A_String := ZERO ;

  BEGIN 	{ Str_Privs }
  FOR P := Cmkrnl TO Security DO
      IF P IN Privs  THEN
	 BEGIN
	 WRITEV (T1,P) ;
	 Str_Compress (T1,T1) ;
	 T2 := T2 + ' ' + SUBSTR (T1,1,4) ;
	 END ;
  Str_Privs := T2 ;
  END ; 	{ Str_Privs }


[GLOBAL] PROCEDURE Format_User (Uic : UIC$TYPE) ;

{ Create a string with the UIC in numeric and identifier formats }

  VAR  Rst : INTEGER := 0 ;

  BEGIN 	{ Format_User }
  Rst := $FAO ('!%U',Uic_Str.LENGTH,%STDESCR Uic_Str.BODY,%IMMED Uic) ;
  Zip (Rst) ;
  Rst := $FAO ('!%I',Id_Str.LENGTH,%STDESCR Id_Str.BODY,%IMMED Uic) ;
  Zip (Rst) ;
  IF Uic_Str = Id_Str  THEN
     Eq_Id_Str := ''
   ELSE
     Eq_Id_Str := ' = ' + Id_Str ;
  END ; 	{ of Format_User }


PROCEDURE Get_System_Info ;

{ Get relevant information about the system environment. }

  VAR	Rst ,
	Arb_Support : INTEGER := 0 ;
	Iosb : Status_Block_Type := ZERO ;
	Item_List : Item_List_Template (6) := ZERO ;

  BEGIN 	{ Get_System_Info }
  Item_List[1].Buffer_Length := 4 ;
  Item_List[1].Item_Code     := SYI$_MAXSYSGROUP ;
  Item_List[1].Buffer_Addr   := IADDRESS (Max_Sys_Group) ;
  Item_List[1].Return_Addr   := 0 ;

  Item_List[2].Buffer_Length := SIZE (System_Rights) ;
  Item_List[2].Item_Code     := SYI$_SYSTEM_RIGHTS ;
  Item_List[2].Buffer_Addr   := IADDRESS (System_Rights) ;
  Item_List[2].Return_Addr   := 0 ;

  Item_List[3].Buffer_Length := 15 ;
  Item_List[3].Item_Code     := SYI$_ARCH_NAME ;
  Item_List[3].Buffer_Addr   := IADDRESS (Vms_Arch.BODY) ;
  Item_List[3].Return_Addr   := IADDRESS (Vms_Arch.LENGTH) ;

  Item_List[4].Buffer_Length := 8 ;
  Item_List[4].Item_Code     := SYI$_VERSION ;
  Item_List[4].Buffer_Addr   := IADDRESS (Vms_Version.BODY) ;
  Item_List[4].Return_Addr   := IADDRESS (Vms_Version.LENGTH) ;

  Item_List[5].Buffer_Length := 4 ;
  Item_List[5].Item_Code     := Syi___Arb_Support ;
  Item_List[5].Buffer_Addr   := IADDRESS (Arb_Support) ;
  Item_List[5].Return_Addr   := 0 ;

  Item_List[6].Terminator    := 0 ;   { Terminate the item list }

  Rst := $GETSYIW (Itmlst := Item_List,
		   Iosb   := Iosb) ;
  Zip (Rst,Iosb[1]) ;

  Arb_Full_Support := Arb_Support = Iss__C_Arb_Full ;

  Rst := STR$UPCASE (%DESCR Vms_Arch,Vms_Arch) ;
  Zip (Rst) ;

  IF (Vms_Arch <> 'VAX') AND
     (Vms_Arch <> 'ALPHA') AND
     (Vms_Arch <> 'IA64') AND
     (Vms_Arch <> 'X86_64') THEN
       $EXIT (Jump__Unsupported) ;	{ Should never be the case! }

  IF Vms_Arch = 'VAX'  THEN
     Architecture := Vax
   ELSE
  IF Vms_Arch = 'IA64'  THEN
     Architecture := Integrity
   ELSE
  IF Vms_Arch = 'X86_64'  THEN
     Architecture := X86 ;

  Vms_Version_Num := SUBSTR (Vms_Version,INDEX(Vms_Version,'.')-1,3) ;

  { The Persona Security Block (PSB) is implemented only on OpenVMS Alpha
    and OpenVMS IA64 from V7.2 onwards. }

  Psb_Available := (Architecture <> Vax) AND (Vms_Version_Num >= '7.2') ;
  END ; 	{ of Get_System_Info }


PROCEDURE Get_Caller_Info ;

{ Get relevant information about the invoker of the program. }

  VAR	Rst : INTEGER := 0 ;
	Iosb : Status_Block_Type := ZERO ;
	Item_List : Item_List_Template (16) := ZERO ;

  BEGIN 	{ Get_Caller_Info }
  Item_List[1].Buffer_Length := 8 ;
  Item_List[1].Item_Code     := JPI$_PROCPRIV ;
  Item_List[1].Buffer_Addr   := IADDRESS (Proc_Perm_Priv) ;
  Item_List[1].Return_Addr   := 0 ;

  Item_List[2].Buffer_Length := 8 ;
  Item_List[2].Item_Code     := JPI$_AUTHPRIV ;
  Item_List[2].Buffer_Addr   := IADDRESS (Proc_Auth_Priv) ;
  Item_List[2].Return_Addr   := 0 ;

  Item_List[3].Buffer_Length := 8 ;
  Item_List[3].Item_Code     := JPI$_CURPRIV ;
  Item_List[3].Buffer_Addr   := IADDRESS (Proc_Cur_Priv) ;
  Item_List[3].Return_Addr   := 0 ;

  Item_List[4].Buffer_Length := 4 ;
  Item_List[4].Item_Code     := JPI$_UIC ;
  Item_List[4].Buffer_Addr   := IADDRESS (Orig_Uic.UIC$L_UIC) ;
  Item_List[4].Return_Addr   := 0 ;

  Item_List[5].Buffer_Length := SIZE (Orig_User) ;
  Item_List[5].Item_Code     := JPI$_USERNAME ;
  Item_List[5].Buffer_Addr   := IADDRESS (Orig_User.BODY) ;
  Item_List[5].Return_Addr   := IADDRESS (Orig_User.LENGTH) ;

  Item_List[6].Buffer_Length := 4 ;
  Item_List[6].Item_Code     := JPI$_PID ;
  Item_List[6].Buffer_Addr   := IADDRESS (Pid) ;
  Item_List[6].Return_Addr   := 0 ;

  Item_List[7].Buffer_Length := 4 ;
  Item_List[7].Item_Code     := JPI$_MASTER_PID ;
  Item_List[7].Buffer_Addr   := IADDRESS (Master_Pid) ;
  Item_List[7].Return_Addr   := 0 ;

  Item_List[8].Buffer_Length := 4 ;
  Item_List[8].Item_Code     := JPI$_PRCCNT ;
  Item_List[8].Buffer_Addr   := IADDRESS (Proc_Cnt) ;
  Item_List[8].Return_Addr   := 0 ;

  Item_List[9].Buffer_Length := SIZE (Port.BODY) ;
  Item_List[9].Item_Code     := JPI$_TT_ACCPORNAM ;
  Item_List[9].Buffer_Addr   := IADDRESS (Port.BODY) ;
  Item_List[9].Return_Addr   := IADDRESS (Port.LENGTH) ;

  Item_List[10].Buffer_Length := SIZE (Terminal.BODY) ;
  Item_List[10].Item_Code     := JPI$_TERMINAL ;
  Item_List[10].Buffer_Addr   := IADDRESS (Terminal.BODY) ;
  Item_List[10].Return_Addr   := IADDRESS (Terminal.LENGTH) ;

  Item_List[11].Buffer_Length := SIZE (Physical_Device.BODY) ;
  Item_List[11].Item_Code     := JPI$_TT_PHYDEVNAM ;
  Item_List[11].Buffer_Addr   := IADDRESS (Physical_Device.BODY) ;
  Item_List[11].Return_Addr   := IADDRESS (Physical_Device.LENGTH) ;

  Item_List[12].Buffer_Length := SIZE (Caller_Rights) ;
  Item_List[12].Item_Code     := JPI$_RIGHTSLIST ;	    { Process + System }
  Item_List[12].Buffer_Addr   := IADDRESS (Caller_Rights) ;
  Item_List[12].Return_Addr   := 0 ;

  Item_List[13].Buffer_Length := 8 ;
  Item_List[13].Item_Code     := JPI$_LOGINTIM ;
  Item_List[13].Buffer_Addr   := IADDRESS (Login_Time) ;
  Item_List[13].Return_Addr   := 0 ;

  Item_List[14].Buffer_Length := 16 ;
  Item_List[14].Item_Code     := JPI$_PRCNAM ;
  Item_List[14].Buffer_Addr   := IADDRESS (Process_Name.BODY) ;
  Item_List[14].Return_Addr   := IADDRESS (Process_Name.LENGTH) ;

  Item_List[15].Buffer_Length := 4 ;
  Item_List[15].Item_Code     := JPI$_JOBPRCCNT ;
  Item_List[15].Buffer_Addr   := IADDRESS (Job_Proc_Cnt) ;
  Item_List[15].Return_Addr   := 0 ;

  Item_List[16].Terminator    := 0 ;	{ Terminate the item list }

  Rst := $GETJPIW (Itmlst := Item_List,
		   Iosb   := Iosb) ;
  Zip (Rst,Iosb[1]) ;

  Rst := STR$TRIM (%DESCR Orig_User,%DESCR Orig_User) ;
  Zip (Rst) ;

  IF (Physical_Device <> '') AND_THEN
     (Physical_Device[1] = '_')  THEN
      Physical_Device := SUBSTR (Physical_Device,2,Physical_Device.LENGTH-1) ;

  IF NOT (Required_Privs <= Proc_Cur_Priv)  THEN	{ Got needed privs? }
     $EXIT (Jump__Nopriv) ;

  Rst := $ASCTIM (Login_Time_Str.LENGTH,Login_Time_Str.BODY,Login_Time) ;
  Zip (Rst) ;
  IF Login_Time_Str[1]	= ' '  THEN
     Login_Time_Str[1] := '0' ;

  { Get the flags }

  Item_List[1].Buffer_Length := 4 ;
  Item_List[1].Item_Code     := UAI$_FLAGS ;
  Item_List[1].Buffer_Addr   := IADDRESS (Caller_Flags) ;
  Item_List[1].Return_Addr   := 0 ;

  Item_List[2].Terminator    := 0 ;	{ Terminate the item list }

  Rst := $GETUAI (Usrnam := %STDESCR SUBSTR (Orig_User,1,Orig_User.LENGTH),
		  Itmlst := Item_List) ;
  Zip (Rst) ;

  Sysprog  := (Orig_Uic.UIC$V_GROUP <= Max_Sys_Group) OR
	      (Setprv IN Proc_Auth_Priv) OR
	      (Setprv IN Proc_Perm_Priv) ;

  Operator := (Oper IN Proc_Auth_Priv) OR
	      (Oper IN Proc_Perm_Priv) ;

  END ; 	{ of Get_Caller_Info }


PROCEDURE Audit_Jump (Ok : BOOLEAN) ;

{ Record who, when, where, how, etc. for auditing purposes. Any problems
  opening the audit file will terminate JUMP. }

  VAR	Rst : UNSIGNED := 0 ;
	Imprint : VARYING [240] OF CHAR := '' ; 	{ 3 lines!! }
	Audit : TEXT ;
	Stamp : PACKED ARRAY [1..23] OF CHAR := '' ;

  BEGIN 	{ Audit_Jump }
  High_Security := TRUE ;
  OPEN (Audit,FILE_NAME:=Audit_Trail,HISTORY:=Old,SHARING:=READWRITE,
	      USER_ACTION:=Special_File_Open,Error:=CONTINUE) ;
  Rst := STATUS (Audit) ;
  IF Rst <> PAS$K_SUCCESS  THEN
     BEGIN
     LIB$SIGNAL (Jump__Badaudit) ;
     LIB$STOP (Condition_Code (Rst)) ;
     END ;

  EXTEND (Audit) ;

  Rst := $ASCTIM (Timbuf := %STDESCR Stamp) ;
  Zip (Rst) ;
  IF Stamp[1] = ' '  THEN  Stamp[1] := '0' ;
  Imprint := SUBSTR (Stamp,1,20) + ' ' + Orig_User ;
  IF Ok  THEN
     BEGIN
     IF (New_User = Orig_User) AND NOT Real_Mccoy  THEN
	BEGIN
	Imprint := Imprint + ' from ' ;
	Format_User (Orig_Uic) ;
	END
      ELSE
	BEGIN
	Imprint := Imprint + '  to  ' ;
	Format_User (New_Uic) ;
	END ;

     IF Alter_Ego AND NOT Transmute  THEN
	Imprint := Imprint + New_User
      ELSE
	Imprint := Imprint + Uic_Str + Eq_Id_Str ;

     IF Alter_Ego  THEN
	Imprint := Imprint + ' *SETUSER*' ;

     IF New_User = Orig_User  THEN
	Format_User (New_Uic) ; 	{ Provide format for display }
     END
   ELSE
     Imprint := Imprint + ' PRIV violation: ' + Command ;

  IF Broken_Chain  THEN
     Imprint := Imprint + ' *CHAINING*' ;

  IF Real_Mccoy  THEN
     BEGIN
     IF Secure_Mode  THEN
	BEGIN
	Imprint := Imprint + ' *SECURE*' ;
	IF Suspect  THEN  Imprint := Imprint + ' ??SUSPECT??' ;
	END
      ELSE
	Imprint := Imprint + ' *EXACT*' ;
     IF Pdev <> ''  THEN  Imprint := Imprint + ' ' + Pdev ;
     END ;

  IF Port <> ''  THEN  Imprint := Imprint + ' ' + Port ;
  IF Terminal <> ''  THEN  Imprint := Imprint + ' ' + Terminal ;
  IF Physical_Device <> ''  THEN  Imprint := Imprint + ' ' + Physical_Device ;

  WRITELN (Audit,Imprint) ;
  CLOSE (Audit) ;

  IF Bugger.Info  THEN
     WRITELN (Dbgfile,'*** Audit record: ',Imprint) ;
  END ; 	{ of Audit_Jump }


PROCEDURE Get_And_Parse_Command ;

{ Get and parse the DCL command line.  Do some basic username checks.
  Evaluate all JUMP logical names. }

  VAR	Rst , Spot : UNSIGNED := 0 ;
	I : INTEGER := 0 ;
	Bool : BOOLEAN := FALSE ;
	Prv : Privilege := Tmpmbx ;
	Jump_Attributes ,
	Str : VARYING [255] OF CHAR := '' ;
	Opclass : Oper_Class := Central ;
	Device_Info : DEV$TYPE := ZERO ;
	Jump_Cld : [EXTERNAL] INTEGER ; { CLD module }

  PROCEDURE Check_Boolean (Boolean_Name : VARYING [Len] OF CHAR ;
			   VAR Boolean_Bit : BOOLEAN) ;

    BEGIN	{ Check_Boolean }
    Rst := CLI$PRESENT (Boolean_Name) ;
    IF (Rst = CLI$_PRESENT) OR (Rst = CLI$_DEFAULTED)  THEN
       Boolean_Bit := TRUE
     ELSE
    IF Rst = CLI$_NEGATED  THEN
       Boolean_Bit := FALSE ;
    END ;	{ Check_Boolean }

  BEGIN 	{ Get_And_Parse_Command }
  Rst := LIB$GET_FOREIGN (%DESCR Command) ;		{ Get the command line }
  Zip (Rst) ;

  Command := 'JUMP ' + Command ;
  Rst := CLI$DCL_PARSE (Command,%REF Jump_Cld,%IMMED LIB$GET_INPUT,
			%IMMED LIB$GET_INPUT,'JUMP> ') ;
  IF (Rst = RMS$_EOF) OR (Rst = CLI$_NOCOMD) OR (NOT ODD (Rst)) THEN  $EXIT ;

  Jump_Version := Jump_Version + ' on OpenVMS ' + Vms_Arch + ' ' +
		  Vms_Version ;
  Str_Compress (Jump_Version,Jump_Version) ;

  Jump_Modules := 'Module build levels: Mn ' + Main_Bl + ', Fd ' +
		  Fiddler_Bl + ', Pk ' + Poker_Bl + ', Ps ' +
		  Persona_Bl + ', Vx ' + Vax_Bl ;
  Str_Compress (Jump_Modules,Jump_Modules) ;

  { If the version is requested, ignore everything else and exit. }

  IF CLI$PRESENT ('VERSION') = CLI$_PRESENT  THEN
     BEGIN
     LIB$SIGNAL (Jump__Version,1,
		 %STDESCR SUBSTR(Jump_Version.BODY,1,Jump_Version.LENGTH)) ;
     LIB$SIGNAL (Jump__Modules,1,
		 %STDESCR SUBSTR(Jump_Modules.BODY,1,Jump_Modules.LENGTH)) ;
     IF NOT Production_Version	THEN  LIB$SIGNAL (Jump__Devnotprd) ;
     $EXIT ;
     END ;

  { If Help is requested, ignore everything else, give help, then exit.
    Help is only available to Sysprogs! }

  IF CLI$PRESENT ('HELP') = CLI$_PRESENT  THEN
     IF NOT Sysprog  THEN
	LIB$SIGNAL (Jump__Nopriv)
      ELSE
	BEGIN
	Rst := CLI$GET_VALUE ('TOPIC',%DESCR Help_Topic) ;    { Might be there }
	IF Rst = CLI$_PRESENT  THEN
	   Str_Compress (Help_Topic,Help_Topic)
	 ELSE
	IF Rst = CLI$_ABSENT  THEN	{ Ignore }
	 ELSE
	   Zip (Rst) ;

	Get_Logical_Name (Lognam  := 'JUMP_HELPLIB',
			  Default := 'SYS_HELP:PUBHELP.HLB',   { SITE-specific }
			  Actual  := Help_Library) ;
	Give_Help ;
	$EXIT ;
	END ;

  Rst := CLI$GET_VALUE ('USERNAME',%DESCR New_User) ;	{ Assume will be there }
  IF Rst = CLI$_ABSENT	THEN
     New_User := Orig_User			 { Default to current username }
   ELSE
  IF NOT ODD (Rst)  THEN
     LIB$SIGNAL (Rst)
   ELSE
     BEGIN
     Rst := STR$TRIM (%DESCR New_User,%DESCR New_User) ;
     Zip (Rst) ;
     END ;

  IF FIND_NONMEMBER (New_User,Symbol) <> 0 THEN
     $EXIT (Jump__Baduser) ;

  Rst := CLI$PRESENT ('EXACT') ;
  Real_Mccoy := (Rst = CLI$_PRESENT) OR (Rst = CLI$_DEFAULTED) ;

  IF Real_Mccoy THEN
     BEGIN
     Get_Logical_Name (Lognam  := 'JUMP_SECURE_MODE',
		       Default := 'FALSE',		       { SITE-specific }
		       Actual  := Str) ;
     Str_Compress (Str,Str) ;
     Rst := STR$UPCASE (%DESCR Str,Str) ;
     Zip (Rst) ;
     System_Secure_Mode := Str = 'TRUE' ;	{ Mandatory Secure_Mode }

     Rst := CLI$PRESENT ('SECURE_MODE') ;	{ Requested Secure_Mode }
     Secure_Mode := (Rst = CLI$_PRESENT) OR (Rst = CLI$_DEFAULTED) ;
     Suspect := Rst = CLI$_NEGATED ;

     Get_Logical_Name (Lognam  := 'JUMP_SECURE_DIR',
		       Default := 'SYS_MANAGER:',	       { SITE-specific }
		       Actual  := Secure_Directory) ;

     Rst := Get_Logical_Name (Lognam  := 'JUMP_USER_DIR',
			      Default := 'SYS$LOGIN:',	       { SITE-specific }
			      Actual  := User_Directory) ;

     { If the System Manager has not explicitly defined the JUMP_USER_DIR
       logical name, allow the user to do so. In this way, EXECUTIVE mode
       definitions will always override definitions in outer access modes. }

     IF Rst = SS$_NOLOGNAM  THEN
	BEGIN
	Rst := Get_Logical_Name (Lognam  := 'JUMP_USER_DIR',
				 Default := 'SYS$LOGIN:',      { SITE-specific }
				 Actual  := User_Directory,
				 Table	 := 'LNM$FILE_DEV',
				 Mode	 := PSL$C_USER) ;
	IF Rst = SS$_NORMAL  THEN
	   Secure_User_Dir := FALSE ;
	END ;

     Get_Logical_Name (Lognam  := 'JUMP_ACME_AGENT_NAME',
		       Default := 'VMS',	       { SITE-specific }
		       Actual  := Str) ;
     Str_Compress (Str,Str) ;
     Rst := STR$UPCASE (%DESCR Str,Str) ;
     Zip (Rst) ;
     Acme_Agent_Name := Str ;

     Rst := CLI$PRESENT ('RECORD') ;
     Record_Session := (Rst = CLI$_PRESENT) OR (Rst = CLI$_DEFAULTED) ;
     Suspect := Suspect OR (Rst = CLI$_NEGATED) ;

     IF Secure_Mode  THEN			{ Requested - set defaults }
	BEGIN
	Notify.All_Bits := 16#Ffffffff ;	{ Full notification }
	Record_Session	:= TRUE ;		{ Required, obviously }
	END ;

     { System_Secure_Mode is handled in Validate_Access.  For now, if
       Secure_Mode is requested, allow other qualifiers to modify the
       notification profile. }

     Rst := CLI$PRESENT ('NOTIFY') ;
     IF Rst = CLI$_PRESENT THEN 	{ Present if not explicitly negated }
	BEGIN
	Rst := CLI$PRESENT ('NOTIFY.BEFORE') ;
	Notify.Before := (Rst = CLI$_PRESENT) OR
			 (Rst = CLI$_DEFAULTED) OR
			 (Secure_Mode AND NOT (Rst = CLI$_NEGATED)) ;

	Rst := CLI$PRESENT ('NOTIFY.AFTER') ;
	Notify.After := (Rst = CLI$_PRESENT) OR
			(Rst = CLI$_DEFAULTED) OR
			(Secure_Mode AND NOT (Rst = CLI$_NEGATED)) ;

	Rst := CLI$PRESENT ('NOTIFY.OPCOM') ;
	Notify.By_Opcom := (Rst = CLI$_PRESENT) OR
			   (Rst = CLI$_DEFAULTED) OR
			   (Secure_Mode AND NOT (Rst = CLI$_NEGATED)) ;

	Rst := CLI$PRESENT ('NOTIFY.MAIL') ;
	Notify.By_Mail := (Rst = CLI$_PRESENT) OR
			  (Rst = CLI$_DEFAULTED) OR
			  (Secure_Mode AND NOT (Rst = CLI$_NEGATED)) ;

	Rst := CLI$PRESENT ('NOTIFY.INCLUDE') ;
	Notify.Include_Log := Record_Session AND
			      Notify.By_Mail AND
			      ((Rst = CLI$_PRESENT) OR
			       (Rst = CLI$_DEFAULTED) OR
			       (Secure_Mode AND NOT (Rst = CLI$_NEGATED))) ;

	Rst := CLI$PRESENT ('NOTIFY.EXIT_ON_MAIL_ERROR') ;
	Notify.Vamoose := (Rst = CLI$_PRESENT) OR
			  (Rst = CLI$_DEFAULTED) OR
			  (Secure_Mode AND Notify.By_Mail) ;

	IF Notify.All_Bits = 0	THEN		{ Default to everything }
	   BEGIN
	   Notify.All_Bits := 16#Ffffffff ;
	   Notify.Include_Log := Record_Session ;
	   END ;
	END
      ELSE
     IF Rst = CLI$_DEFAULTED  THEN
	BEGIN
	Notify.All_Bits    := 16#Ffffffff ;	{ Full notification, almost? }
	Notify.Include_Log := Record_Session ;
	END ;

     Rst := CLI$PRESENT ('ESCAPE_CHARACTER') ;
     IF Rst = CLI$_PRESENT THEN 		{ Turn into Control character }
	BEGIN
	Rst := CLI$GET_VALUE ('ESCAPE_CHARACTER',%DESCR Str) ;
	IF Rst = CLI$_ABSENT  THEN
	   Ripcord := CHR (ORD(Ripcord)-64)		{ Use default }
	 ELSE
	IF NOT ODD (Rst)  THEN
	   LIB$SIGNAL (Rst)
	 ELSE
	   BEGIN
	   Rst := STR$UPCASE (%DESCR Str,Str) ;
	   Zip (Rst) ;
	   IF (Str.LENGTH <> 1) OR_ELSE NOT (Str[1] IN Esc_Chars)  THEN
	      LIB$SIGNAL (Jump__Invescchar)
	    ELSE
	      Ripcord := CHR (ORD(Str[1])-64) ;
	   END ;
	Escape_Hatch := TRUE ;
	END ;

     Rst := CLI$PRESENT ('COMMAND') ;
     IF Rst = CLI$_PRESENT THEN
	BEGIN
	Rst := CLI$GET_VALUE ('COMMAND',%DESCR Str) ;
	IF NOT ODD (Rst)  THEN
	   LIB$SIGNAL (Rst)
	 ELSE
	   BEGIN
	   Single_Cmd_Mode := TRUE ;
	   Str_Compress (Str,Str) ;
	   IF Str[1] = '"'  THEN
	      BEGIN
	      Str := SUBSTR (Str,2,Str.LENGTH-1) ;
	      IF Str[Str.LENGTH] = '"'	THEN
		 Str := SUBSTR (Str,1,Str.LENGTH-1) ;
	      Str_Compress (Str,Str) ;
	      END ;
	   IF Str[1] = '$'  THEN
	      Str := SUBSTR (Str,2,Str.LENGTH-1) ;
	   Str_Compress (Str,Str) ;
	   Single_Cmd := Str ;
	   END ;
	END ;

     Rst := CLI$PRESENT ('SELECT') ;
     IF Rst = CLI$_PRESENT THEN
	BEGIN
	Rst := CLI$GET_VALUE ('SELECT',%DESCR Str) ;
	IF NOT ODD (Rst)  THEN
	   LIB$SIGNAL (Rst)
	 ELSE
	   READV (Str,Scm_Selection) ;
	END ;
     END ;	{ of IF Real_Mccoy }

  Get_Logical_Name (Lognam  := 'JUMP_NOTIFY_MAILLIST',
		    Default := 'SYSTEM',		       { SITE-specific }
		    Actual  := Notify_Maillist) ;
  Str_Compress (Notify_Maillist,Notify_Maillist,TRUE) ;

  Check_Boolean ('SETUSER',Alter_Ego) ;
  Transmute := NOT (Alter_Ego OR Real_Mccoy) ;

  Rst := CLI$PRESENT ('ALL') ;
  Alter_Ego := Alter_Ego OR (Rst = CLI$_PRESENT) OR (Rst = CLI$_DEFAULTED) ;

  Check_Boolean ('OVERRIDE_UAF',Figment) ;
  Check_Boolean ('UPDATE_UAF',Stamp_Uaf) ;
  Check_Boolean ('LOG',Log) ;
  Check_Boolean ('AUDIT',Auditing) ;

  Get_Logical_Name (Lognam  := 'JUMP_AUDIT_TRAIL',
		    Default := 'SYS_MANAGER:JUMP_AUDIT.DAT',   { SITE-specific }
		    Actual  := Audit_Trail) ;

  Get_Logical_Name (Lognam  := 'JUMP_ACCESS_LIST',
		    Default := 'SYS_MANAGER:JUMP_ACCESS.DAT',  { SITE-specific }
		    Actual  := Access_List) ;

  Get_Logical_Name (Lognam  := 'JUMP_DOUBLE_CHECK',
		    Default := 'TRUE',			       { SITE-specific }
		    Actual  := Str) ;
  Str_Compress (Str,Str) ;
  Rst := STR$UPCASE (%DESCR Str,Str) ;
  Zip (Rst) ;
  Double_Check := Str = 'TRUE' ;

  { If the current process is a JUMP pseudo-terminal, check for any attributes.
    Prevent chaining, if necessary. }

  Get_Logical_Name (Lognam  := 'JUMP_' + HEX (Pid,8,8),     { Process-specific }
		    Default := 'NONE',			       { SITE-specific }
		    Actual  := Jump_Attributes) ;
  IF (INDEX (Jump_Attributes,'NOCHAIN ') > 0) AND	{ Chaining not allowed }
     (New_User <> Orig_User)  THEN             { ... and it's a real chain ... }
       BEGIN
       Broken_Chain := TRUE ;
       IF Auditing  THEN
	  Audit_Jump (FALSE) ;
       $EXIT (Jump__Nochain) ;                  { ... c'est bon soir ma chere! }
       END ;

  { Allow an EXACT JUMP to chain? }

  Get_Logical_Name (Lognam  := 'JUMP_CHAIN',
		    Default := 'TRUE',			       { SITE-specific }
		    Actual  := Str) ;
  Str_Compress (Str,Str) ;
  Rst := STR$UPCASE (%DESCR Str,Str) ;
  Zip (Rst) ;
  Chain := Str = 'TRUE' ;

  Get_Logical_Name (Lognam  := 'JUMP_ALLOW_DEBUG_'+Orig_User,
		    Default := 'FALSE', 		{ SITE-specific }
		    Actual  := Str) ;
  Str_Compress (Str,Str) ;
  Rst := STR$UPCASE (%DESCR Str,Str) ;
  Zip (Rst) ;
  Allow_Debug := Str = 'TRUE' ;

  Get_Logical_Name (Lognam  := 'JUMP_USER_DEBUG_OPCOM',
		    Default := 'TRUE',		{ SITE-specific }
		    Actual  := Str) ;
  Str_Compress (Str,Str) ;
  Rst := STR$UPCASE (%DESCR Str,Str) ;
  Zip (Rst) ;
  User_Debug_Opcom := Str = 'TRUE' ;

  Get_Logical_Name (Lognam  := 'JUMP_SELF',
		    Default := 'TRUE',			{ SITE-specific }
		    Actual  := Str) ;
  Str_Compress (Str,Str) ;
  Rst := STR$UPCASE (%DESCR Str,Str) ;
  Zip (Rst) ;
  Narcissus := Str = 'TRUE' ;

  Get_Logical_Name (Lognam  := 'JUMP_MINOR_PRIVS',
		    Default := 'NETMBX TMPMBX', 	{ SITE-specific }
		    Actual  := Str) ;
  Str_Compress (Str,Str) ;
  REPEAT
    READV (Str,Prv,Error:=CONTINUE) ;
    IF STATUSV <> 0  THEN
       $EXIT (Jump__Badprivset)
     ELSE
       BEGIN
       Minor_Privs := Minor_Privs + [Prv] ;
       Spot := INDEX (Str,' ') ;
       IF Spot = 0  THEN
	  Str := ''
	ELSE
	  Str := SUBSTR (Str,Spot+1,Str.LENGTH-Spot) ;
       END ;
  UNTIL Str = '' ;

  Get_Logical_Name (Lognam  := 'JUMP_OPER_CLASSES',
		    Default := 'CENTRAL',		{ SITE-specific }
		    Actual  := Str) ;
  Str_Compress (Str,Str) ;
  Rst := STR$UPCASE (%DESCR Str,Str) ;
  Zip (Rst) ;
  IF Str = 'SECURITY'  THEN		{ Do nothing - SECURITY only required }
   ELSE
     BEGIN
     Spot := INDEX (Str,'SECURITY') ;
     IF Spot > 0  THEN
	BEGIN
	IF Spot = 1  THEN			{ First up }
	   Str := SUBSTR (Str,10,Str.LENGTH-9)
	 ELSE
	IF Spot+7 = Str.LENGTH	THEN		{ Last one }
	   Str := SUBSTR (Str,1,Str.LENGTH-9)
	 ELSE
	   Str := SUBSTR (Str,1,Spot-1) +	{ Sandwiched }
		  SUBSTR (Str,Spot+9,Str.LENGTH-Spot-8) ;
	Str_Compress (Str,Str) ;
	END ;

     REPEAT
       READV (Str,Opclass,Error:=CONTINUE) ;
       IF STATUSV <> 0	THEN
	  $EXIT (Jump__Badoperclass)
	ELSE
	  BEGIN
	  Oper_Classes := Oper_Classes + [Opclass] ;
	  Spot := INDEX (Str,' ') ;
	  IF Spot = 0  THEN
	     Str := ''
	   ELSE
	     Str := SUBSTR (Str,Spot+1,Str.LENGTH-Spot) ;
	  END ;
     UNTIL Str = '' ;
     END ;

  Oper_Classes := Oper_Classes + [Opc_Security] ;      { SECURITY is mandatory }
  Oper_Class_Mask := Oper_Classes::UNSIGNED ;

  Check_Boolean ('DEBUG',Debug) ;
  IF Debug  THEN
     IF NOT (Sysprog OR Allow_Debug)	THEN
	LIB$SIGNAL (Jump__Nopriv)
      ELSE
	BEGIN
	IF Allow_Debug	THEN
	   BEGIN
	   LIB$SIGNAL (Jump__Debugon) ;
	   IF User_Debug_Opcom	THEN
	      Oprmsg ('JUMP user '+Orig_User+' is enabled for DEBUG',
		      Oper_Class_Mask) ;
	   END ;

	Rst := CLI$PRESENT ('DEBUG.ALL') ;
	IF Rst = CLI$_PRESENT  THEN
	   Bugger.All_Bits := 16#Ffffffff
	 ELSE
	IF Rst = CLI$_NEGATED  THEN
	   Bugger.All_Bits := 0 ;

	Bool := Bugger.Info ;		{ General Information }
	Check_Boolean ('DEBUG.INFO',Bool) ;
	Bugger.Info := Bool ;

	{ Use Bugger.Dn to diagnose by tracing execution or showing results }

	Bool := Bugger.D1 ;		{ TRACE Pchan/Rchan }
	Check_Boolean ('DEBUG.D1',Bool) ;
	Bugger.D1 := Bool ;

	Bool := Bugger.D2 ;		{ TRACE Parse Number List }
	Check_Boolean ('DEBUG.D2',Bool) ;
	Bugger.D2 := Bool ;

	Bool := Bugger.D3 ;		{ SHOW Parse Number List results }
	Check_Boolean ('DEBUG.D3',Bool) ;
	Bugger.D3 := Bool ;

	Bool := Bugger.D4 ;		{ TRACE Pseudoterm creation }
	Check_Boolean ('DEBUG.D4',Bool) ;
	Bugger.D4 := Bool ;

	Bool := Bugger.D5 ;
	Check_Boolean ('DEBUG.D5',Bool) ;
	Bugger.D5 := Bool ;

	Bool := Bugger.D6 ;
	Check_Boolean ('DEBUG.D6',Bool) ;
	Bugger.D6 := Bool ;

	Bool := Bugger.D7 ;
	Check_Boolean ('DEBUG.D7',Bool) ;
	Bugger.D7 := Bool ;

	Bool := Bugger.D8 ;
	Check_Boolean ('DEBUG.D8',Bool) ;
	Bugger.D8 := Bool ;

	Bool := Bugger.D9 ;
	Check_Boolean ('DEBUG.D9',Bool) ;
	Bugger.D9 := Bool ;
	IF Bugger.D9  THEN
	   BEGIN
	   WRITELN ('1 Pch, 2 Num, 3 Res, 4 Ptm, 9 Sho. ',
		    '1 Non, 2 Per, 3 GEH, 6 TKO, 9 Abt.') ;
	   $EXIT ;
	   END ;

	{ Use Bugger.Xn to change execution path or behaviour }

	Bool := Bugger.X1 ;		{ Force NOT Use Persona }
	Check_Boolean ('DEBUG.X1',Bool) ;
	Bugger.X1 := Bool ;

	Bool := Bugger.X2 ;		{ Force Use Persona }
	Check_Boolean ('DEBUG.X2',Bool) ;
	Bugger.X2 := Bool ;

	Bool := Bugger.X3 ;		{ Enable General Error Handler }
	Check_Boolean ('DEBUG.X3',Bool) ;
	Bugger.X3 := Bool ;

	Bool := Bugger.X4 ;
	Check_Boolean ('DEBUG.X4',Bool) ;
	Bugger.X4 := Bool ;

	Bool := Bugger.X5 ;
	Check_Boolean ('DEBUG.X5',Bool) ;
	Bugger.X5 := Bool ;

	Bool := Bugger.X6 ;		{ Deliberately crash JUMP }
	Check_Boolean ('DEBUG.X6',Bool) ;
	Bugger.X6 := Bool AND (CLI$PRESENT ('DEBUG.ALL') <> CLI$_PRESENT) ;

	Bool := Bugger.X7 ;
	Check_Boolean ('DEBUG.X7',Bool) ;
	Bugger.X7 := Bool ;

	Bool := Bugger.X8 ;
	Check_Boolean ('DEBUG.X8',Bool) ;
	Bugger.X8 := Bool ;

	Bool := Bugger.X9 ;		{ ABORT before pseudoterm is created }
	Check_Boolean ('DEBUG.X9',Bool) ;
	Bugger.X9 := Bool ;

	Dbgfile_Name := 'SYS$OUTPUT' ;		{ Default output }
	IF CLI$PRESENT ('DEBUG.FILE') = CLI$_PRESENT  THEN
	   BEGIN
	   Rst := CLI$GET_VALUE ('DEBUG.FILE',%DESCR Dbgfile_Name) ;
	   IF Rst = CLI$_ABSENT  THEN  { Create default filename }
	      BEGIN
	      GETTIMESTAMP (Time_Now) ;
	      Time_Str := DATE (Time_Now) + ' ' + TIME (Time_Now) ;
	      IF Time_Str[1]  = ' '  THEN
		 Time_Str[1] := '0' ;
	      WRITEV (Dbgfile_Name,User_Directory,
			 'JUMP_',
			 Orig_User,
			 '-',
			 New_User,
			 '_DEBUG.',
			 DEC (Time_Now.Year,4,4),
			 DEC (Time_Now.Month,2,2),
			 DEC (Time_Now.Day,2,2),
			 '_',
			 DEC (Time_Now.Hour,2,2),
			 DEC (Time_Now.Minute,2,2),
			 DEC (Time_Now.Second,2,2) ) ;

	      END
	    ELSE
	      Zip (Rst) ;
	   LIB$SIGNAL (Jump__Debugfile,1,
		       %STDESCR SUBSTR (Dbgfile_Name,1,Dbgfile_Name.LENGTH)) ;
	   END ;

	Open_New_File := TRUE ;
	OPEN (Dbgfile,FILE_NAME:=Dbgfile_Name,HISTORY:=NEW,SHARING:=READWRITE,
		      USER_ACTION:=Special_File_Open,Error:=CONTINUE) ;
	Rst := STATUS (Dbgfile) ;
	IF Rst <> PAS$K_SUCCESS  THEN
	   BEGIN
	   LIB$SIGNAL (Jump__Baddbgfil) ;
	   LIB$STOP (Condition_Code (Rst)) ;
	   END ;
	REWRITE (Dbgfile) ;

	Dbgfile_Name := Open_File_Name ;
	Device_Info::UNSIGNED := Open_File_Fab.FAB$L_DEV ;
	Debug_To_Terminal := Device_Info.DEV$V_TRM ;

	IF Bugger.Info	THEN
	   BEGIN
	   WRITELN (Dbgfile,'*** ',Jump_Version,' ***') ;
	   WRITELN (Dbgfile,'*** ',Jump_Modules,' ***') ;
	   WRITELN (Dbgfile,'*** DEBUG Mode ***') ;
	   IF Bugger.X3  THEN
	      WRITELN (Dbgfile,'*** WARNING! ',
		       'Condition value signal interception enabled! ***') ;
	   IF NOT Production_Version  THEN
	      WRITELN (Dbgfile,'*** DEVELOPMENT VERSION ***') ;
	   IF NOT Debug_To_Terminal  THEN
	      WRITELN (Dbgfile,'*** Debug File: ',Dbgfile_Name) ;
	   WRITELN (Dbgfile,'*** Audit Trail: ',Audit_Trail) ;
	   WRITELN (Dbgfile,'*** Access List: ',Access_List) ;
	   WRITELN (Dbgfile,'*** Command: ',Command) ;
	   IF Real_Mccoy  THEN
	      WRITELN (Dbgfile,'*** Calling Mode: EXACT')
	    ELSE
	      WRITELN (Dbgfile,'*** Calling Mode: NON-EXACT') ;
	   WRITELN (Dbgfile,'*** Caller Username: ',Orig_User) ;
	   Format_User (Orig_Uic) ;
	   WRITELN (Dbgfile,'*** Caller UIC: ',Id_Str) ;
	   WRITELN (Dbgfile,'*** Caller PID: ',HEX(Pid,8,8)) ;
	   WRITELN (Dbgfile,'*** Caller Process Name: ',Process_Name) ;
	   WRITELN (Dbgfile,'*** Caller Cur  privs:',Str_Privs(Proc_Cur_Priv)) ;
	   WRITELN (Dbgfile,'*** Caller Def  privs:',Str_Privs(Proc_Perm_Priv));
	   WRITELN (Dbgfile,'*** Caller Auth privs:',Str_Privs(Proc_Auth_Priv));

	   WRITE (Dbgfile,'*** Caller Rights: ') ;
	   I := 1 ;
	   WHILE Caller_Rights[I].L0 <> 0  DO
	       BEGIN
	       Rst := $FAO ('!%I',Id_Str.LENGTH,%STDESCR Id_Str.BODY,
			    %IMMED Caller_Rights[I].L0) ;
	       Zip (Rst) ;
	       WRITE (Dbgfile,Id_Str,' ') ;
	       I := I + 1 ;
	       IF (I REM 10 = 0) AND (Caller_Rights[I].L0 <> 0)  THEN
		  BEGIN
		  WRITELN (Dbgfile) ;
		  WRITE   (Dbgfile,'... Caller Rights: ') ;
		  END ;
	       END ;
	   WRITELN (Dbgfile) ;

	   WRITELN (Dbgfile,'*** PSB Available:',Psb_Available) ;
	   IF Psb_Available  THEN
	      BEGIN
	      Check_Extauth_Flags (Caller_Flags) ;
	      WRITELN (Dbgfile,'*** Caller EXTAUTH:',Extauth_Flag) ;
	      WRITELN (Dbgfile,'*** Caller VMSAUTH:',Vmsauth_Flag) ;
	      IF Acme_Agent_Name = 'VMS'  THEN
		 WRITELN (Dbgfile,'*** ACME Agent: VMS only')
	       ELSE
	      IF Acme_Agent_Name <> ''	THEN
		 WRITELN (Dbgfile,'*** ACME Agents: VMS and ',Acme_Agent_Name) ;
	      END ;
	   WRITELN (Dbgfile,'*** Max Sys Grp (Octal): ',OCT(Max_Sys_Group,3)) ;
	   IF Real_Mccoy AND System_Secure_Mode  THEN
	      WRITELN (Dbgfile,'*** System Secure Mode') ;
	   WRITELN (Dbgfile,'*** CHAIN:',Chain) ;
	   WRITELN (Dbgfile,'*** Caller Attributes: ',Jump_Attributes) ;
	   WRITELN (Dbgfile,'*** DOUBLE CHECK:',Double_Check) ;
	   END ;
	END ;

  END ; { of Get_And_Parse_Command }


FUNCTION Get_Target_Info : BOOLEAN ;

{ Get the required information for the particular new user from the UAF. }

  VAR	I , J : INTEGER := 0 ;
	Rst , Ctx : UNSIGNED := 0 ;
	Holder : $UQUAD := ZERO ;
	Item_List : Item_List_Template (7) := ZERO ;

  BEGIN 	{ Get_Target_Info }
  Item_List[1].Buffer_Length := 8 ;
  Item_List[1].Item_Code     := UAI$_DEF_PRIV ;
  Item_List[1].Buffer_Addr   := IADDRESS (Def_Priv) ;
  Item_List[1].Return_Addr   := 0 ;

  Item_List[2].Buffer_Length := 8 ;
  Item_List[2].Item_Code     := UAI$_PRIV ;
  Item_List[2].Buffer_Addr   := IADDRESS (Auth_Priv) ;
  Item_List[2].Return_Addr   := 0 ;

  Item_List[3].Buffer_Length := 4 ;
  Item_List[3].Item_Code     := UAI$_UIC ;
  Item_List[3].Buffer_Addr   := IADDRESS (New_Uic.UIC$L_UIC) ;
  Item_List[3].Return_Addr   := 0 ;

  Item_List[4].Buffer_Length := 4 ;
  Item_List[4].Item_Code     := UAI$_FLAGS ;
  Item_List[4].Buffer_Addr   := IADDRESS (Target_Flags) ;
  Item_List[4].Return_Addr   := 0 ;

  Item_List[5].Buffer_Length := SIZE (Def_Dir) ;
  Item_List[5].Item_Code     := UAI$_DEFDIR ;
  Item_List[5].Buffer_Addr   := IADDRESS (Def_Dir.BODY) ;
  Item_List[5].Return_Addr   := 0 ;

  Item_List[6].Buffer_Length := SIZE (Def_Dev) ;
  Item_List[6].Item_Code     := UAI$_DEFDEV ;
  Item_List[6].Buffer_Addr   := IADDRESS (Def_Dev.BODY) ;
  Item_List[6].Return_Addr   := 0 ;

  Item_List[7].Terminator    := 0 ;	{ Terminate the item list }

  Rst := $GETUAI (Usrnam := %STDESCR SUBSTR (New_User,1,New_User.LENGTH),
		  Itmlst := Item_List) ;

  Get_Target_Info := ODD (Rst) ;

  IF NOT ODD (Rst)  THEN
     BEGIN
     IF Rst <> RMS$_RNF  THEN
	LIB$SIGNAL (Rst) ;
     END
   ELSE
     BEGIN
     Def_Dir.LENGTH := INT (Def_Dir.BODY[1]) ;
     Def_Dev.LENGTH := INT (Def_Dev.BODY[1]) ;
     Def_Dir.BODY := SUBSTR (Def_Dir.BODY,2,Def_Dir.LENGTH) ;
     Def_Dev.BODY := SUBSTR (Def_Dev.BODY,2,Def_Dev.LENGTH) ;

     { Get the Rights IDs held by the target user }

     Holder.L0 := New_Uic.UIC$L_UIC ;
     REPEAT
       I := I + 1 ;
       Rst := $FIND_HELD (Holder := Holder,
			  Id	 := Target_Rights[I].L0,
			  Contxt := Ctx) ;
       IF (Rst <> SS$_NOSUCHID) AND NOT ODD (Rst)  THEN
	  LIB$SIGNAL (Rst) ;
     UNTIL (Rst = SS$_NOSUCHID) OR (I = Rightsize) ;

     I := I - 1 ;
     J := 1 ;
     WHILE System_Rights[J].L0 > 0  DO
       BEGIN
       Target_Rights[I+J].L0 := System_Rights[J].L0 ;
       J := J + 1 ;
       END ;
     END ;

  IF Bugger.Info  THEN
     BEGIN
     WRITELN (Dbgfile,'*** Target Username: ',New_User) ;
     Format_User (New_Uic) ;
     WRITELN (Dbgfile,'*** Target UIC: ',Id_Str) ;
     WRITELN (Dbgfile,'*** Target Def  privs:',Str_Privs(Def_Priv)) ;
     WRITELN (Dbgfile,'*** Target Auth privs:',Str_Privs(Auth_Priv)) ;

     WRITE (Dbgfile,'*** Target Rights: ') ;
     FOR I := 1 TO J-1	DO
	 BEGIN
	 Rst := $FAO ('!%I',Id_Str.LENGTH,%STDESCR Id_Str.BODY,
		      %IMMED Target_Rights[I].L0) ;
	 Zip (Rst) ;
	 WRITE (Dbgfile,Id_Str,' ') ;
	 IF (I REM 10 = 9) AND (Target_Rights[I+1].L0 <> 0)  THEN
	    BEGIN
	    WRITELN (Dbgfile) ;
	    WRITE   (Dbgfile,'... Target Rights: ') ;
	    END ;
	 END ;
     WRITELN (Dbgfile) ;

     IF Psb_Available  THEN
	BEGIN
	Check_Extauth_Flags (Target_Flags) ;
	WRITELN (Dbgfile,'*** Target EXTAUTH:',Extauth_Flag) ;
	WRITELN (Dbgfile,'*** Target VMSAUTH:',Vmsauth_Flag) ;
	END ;
     END ;

  END ; 	{ of Get_Target_Info }


FUNCTION Valid_Username (User : VARYING [Len] OF CHAR) : BOOLEAN ;

{ Determine if the string corresponds to a valid username in the UAF. }

  VAR	Rst , Flags : UNSIGNED := 0 ;
	Item_List : Item_List_Template (2) := ZERO ;

  BEGIN 	{ Valid_Username }
  Item_List[1].Buffer_Length := 4 ;
  Item_List[1].Item_Code     := UAI$_FLAGS ;
  Item_List[1].Buffer_Addr   := IADDRESS (Flags) ;	{ Anything will do }
  Item_List[1].Return_Addr   := 0 ;

  Item_List[2].Terminator    := 0 ;	{ Terminate the item list }

  Rst := $GETUAI (Usrnam := %STDESCR SUBSTR (User,1,Len),
		  Itmlst := Item_List) ;

  Valid_Username := ODD (Rst) ;

  IF NOT ODD (Rst) AND_THEN (Rst <> RMS$_RNF)  THEN
     LIB$SIGNAL (Rst) ;
  END ; 	{ of Valid_Username }


FUNCTION Parse_Ident (Ident_Str : VARYING [Len] OF CHAR ;
		      VAR Ident_Val : UIC$TYPE ;
		      VAR Parse_Result : UNSIGNED ;
		      Req_Type : UNSIGNED := 0) : BOOLEAN ;

  { Use LIB$TABLE_PARSE to parse an identifier string.	This neatly handles
    all parsing issues associated with UICs and identifiers. }

  CONST

  No_Req  = 0 ; 	{ No specific requirements - any valid value is OK }
  Req_Rid = 1 ; 	{ Must be a Rights ID }
  Req_Uic = 2 ; 	{ Must be a UIC }
  Req_Uic_Nowild = 3 ;	{ Must be a non-wildcarded UIC }
  Req_Uic_Wild	 = 4 ;	{ Must be a wildcarded UIC }

  VAR

  Rst : UNSIGNED := 0 ;

  State_Tbl ,				      { TABLE_PARSE table - see source }
  Key_Tbl   : [EXTERNAL,VALUE] UNSIGNED ;     { TABLE_PARSE table - see source }
  Arg_Block : [VOLATILE] TPA$TYPE := ZERO ;   { TABLE_PARSE argument block }

  BEGIN 	{ Parse_Ident }
  Parse_Ident := FALSE ;

  Ident_Val.UIC$L_UIC := 0 ;			{ Return null if error }
  Ident_Val.UIC$V_FORMAT := 1 ; 		{ Invalid format value }

  Arg_Block.TPA$L_COUNT     := TPA$K_COUNT0 ;
  Arg_Block.TPA$L_STRINGCNT := Len ;
  Arg_Block.TPA$L_STRINGPTR := IADDRESS (Ident_Str.BODY) ;

  Rst := LIB$TABLE_PARSE (%REF Arg_Block,
			  %IMMED State_Tbl,	{ See JUMP MACRO source }
			  %IMMED Key_Tbl) ;	{ See JUMP MACRO source }

  Parse_Result := Rst ;

  IF Rst <> LIB$_SYNTAXERR  THEN
     IF NOT ODD (Rst)  THEN
	LIB$SIGNAL (Rst)
      ELSE
	BEGIN
	Ident_Val.UIC$L_UIC := Arg_Block.TPA$L_NUMBER ;
	IF Req_Type = No_Req  THEN
	   Parse_Ident := TRUE
	 ELSE
	IF (Ident_Val.UIC$V_FORMAT = UIC$K_ID_FORMAT) AND
	   (Req_Type = Req_Rid)  THEN
	   Parse_Ident := TRUE
	 ELSE
	IF Ident_Val.UIC$V_FORMAT = UIC$K_UIC_FORMAT  THEN
	   IF (Req_Type  = Req_Uic) OR
	      ((Req_Type = Req_Uic_Nowild) AND
	       (Ident_Val.UIC$V_GROUP  <> UIC$K_WILD_GROUP) AND
	       (Ident_Val.UIC$V_MEMBER <> UIC$K_WILD_MEMBER)) OR
	      ((Req_Type = Req_Uic_Wild) AND
		((Ident_Val.UIC$V_GROUP  = UIC$K_WILD_GROUP) OR
		 (Ident_Val.UIC$V_MEMBER = UIC$K_WILD_MEMBER)))  THEN
	      Parse_Ident := TRUE ;
	END ;
  END ; 	{ of Parse_Ident }


FUNCTION Match_Uic (Id1 , Id2 : UIC$TYPE ;
		    Exact   : BOOLEAN := FALSE ;
		    Ordered : BOOLEAN := FALSE) : BOOLEAN ;

{ Determine if a pair of UICs match, either uniquely if "EXACT", or as
  wildcarded UICs.  If "ORDERED", Id1 must be a "subset" of Id2. }

  BEGIN 	{ Match_UIC }
  Match_Uic := FALSE ;

  IF Id1.UIC$L_UIC = Id2.UIC$L_UIC  THEN		{ Exact match }
     Match_Uic := TRUE
   ELSE
  IF NOT Exact	THEN					{ Wildcarded forms }
     IF Ordered  THEN
	Match_Uic := ((Id2.UIC$V_MEMBER = UIC$K_WILD_MEMBER) OR
		      (Id1.UIC$V_MEMBER = Id2.UIC$V_MEMBER)) AND
		     ((Id2.UIC$V_GROUP	= UIC$K_WILD_GROUP) OR
		      (Id1.UIC$V_GROUP	= Id2.UIC$V_GROUP))
      ELSE
	Match_Uic := ((Id1.UIC$V_MEMBER = UIC$K_WILD_MEMBER) OR
		      (Id2.UIC$V_MEMBER = UIC$K_WILD_MEMBER) OR
		      (Id1.UIC$V_MEMBER = Id2.UIC$V_MEMBER)) AND
		     ((Id1.UIC$V_GROUP	= UIC$K_WILD_GROUP) OR
		      (Id2.UIC$V_GROUP	= UIC$K_WILD_GROUP) OR
		      (Id1.UIC$V_GROUP	= Id2.UIC$V_GROUP)) ;
  END ; 	{ of Match_UIC }


FUNCTION Match_List (Candidate	: VARYING [L1] OF CHAR ;
		     Targets : VARYING [L2] OF CHAR ;
		     Candidate_Uic : UIC$TYPE ;
		     Candidate_Rights : Rights_Array) : UNSIGNED ;

  { Determine if a given candidate string, which may be either a (wildcarded)
    username, a (wildcarded) UIC, or a rights identifier, is included in a
    list of targets which may be of any of these same types. }

  TYPE	Style_Type = (Unknown,Wild_S,User_S,Uic_S,Ident_S) ;

  VAR	Rst , Spot : UNSIGNED := 0 ;
	Negated , Matched : BOOLEAN := FALSE ;
	Id_Style : Style_Type := Unknown ;
	Dancer : VARYING [100] OF CHAR := '' ;
	Dancer_Id : UIC$TYPE := ZERO ;

  BEGIN 	{ Match_List }
  Match_List := CLI$_ABSENT ;

  WHILE Targets <> ''  DO
    BEGIN
    IF Targets[1] = '!'  THEN			{ Negated }
       BEGIN
       IF Targets.LENGTH = 1  THEN		{ Last character??? }
	  $EXIT (Jump__Baddata)
	ELSE
	  Targets := SUBSTR (Targets,2,Targets.LENGTH-1) ;
       Negated := TRUE ;
       END
     ELSE
       Negated := FALSE ;

    IF Targets[1] IN Symbol+Wildcard  THEN
       BEGIN
       Spot := FIND_NONMEMBER (Targets,Symbol+Wildcard) ;
       IF Spot = 0  THEN
	  BEGIN
	  Dancer  := Targets ;
	  Targets := '' ;
	  END
	ELSE
       IF Targets[Spot] = ','  THEN
	  BEGIN
	  Spot := Spot - 1 ;
	  Dancer := SUBSTR (Targets,1,Spot) ;
	  END
	ELSE
	  $EXIT (Jump__Baddata) ;
       IF FIND_MEMBER (Dancer,Wildcard) > 0  THEN
	  Id_Style := Wild_S
	ELSE
	  Id_Style := User_S ;	{ User or ID?  Assume user for now. }
       END
     ELSE
    IF Targets[1] IN Uic_Left  THEN
       BEGIN
       Spot := FIND_MEMBER (Targets,Uic_Right) ;
       IF Spot = 0  THEN
	  $EXIT (Jump__Baddata) ;
       Dancer := SUBSTR (Targets,1,Spot) ;
       IF ((Dancer[1] = '[') AND (Dancer[Dancer.LENGTH] <> ']')) OR
	  ((Dancer[1] = '<') AND (Dancer[Dancer.LENGTH] <> '>'))  THEN
	  $EXIT (Jump__Baddata) ;
       Id_Style := Uic_S ;
       END
     ELSE
       $EXIT (Jump__Baddata) ;

    IF Spot < Targets.LENGTH  THEN
       Targets := SUBSTR (Targets,Spot+1,Targets.LENGTH-Spot)
     ELSE
       Targets := '' ;
    IF (Targets.LENGTH > 1) AND_THEN (Targets[1] = ',')  THEN
	Targets := SUBSTR (Targets,2,Targets.LENGTH-1) ;

    IF Id_Style = Unknown  THEN 		  { Huh??? Should never happen }
       $EXIT (Jump__Baddata)
     ELSE
    IF Id_Style = Wild_S  THEN				   { Wildcard username }
       BEGIN
       Rst := STR$MATCH_WILD (Candidate,Dancer) ;
       IF Rst = STR$_MATCH  THEN
	  Matched := TRUE
	ELSE
       IF NOT ODD (Rst) AND_THEN (Rst <> STR$_NOMATCH)	THEN
	  LIB$SIGNAL (Rst) ;
       END
     ELSE
    IF (Id_Style = User_S) AND_THEN			      { Exact username }
       (Candidate = Dancer)  THEN
	  Matched := TRUE
     ELSE					 { Another username, UIC or ID }
    IF NOT Parse_Ident (Dancer,Dancer_Id,Rst) AND_THEN
       (((Rst  = LIB$_SYNTAXERR) AND (Id_Style = Uic_S)) OR   { Bad UIC syntax }
	((Rst <> LIB$_SYNTAXERR) AND NOT ODD (Rst)))  THEN { Parse_Ident error }
	  $EXIT (Jump__Baddata)
     ELSE
       BEGIN

      { If we reach here, it's a UIC, a UIC ID or a rights ID that has parsed
	successfully to an existing identifier. Ignore UIC IDs that are also
	usernames to eliminate ambiguity -- usernames always take precedence. }

       IF (Id_Style = Uic_S) OR_ELSE
	  (NOT Valid_Username (Dancer))  THEN
	  IF Dancer_Id.UIC$V_FORMAT = UIC$K_ID_FORMAT  THEN
	     Id_Style := Ident_S
	   ELSE
	  IF Dancer_Id.UIC$V_FORMAT = UIC$K_UIC_FORMAT	THEN
	     Id_Style := Uic_S ;

       IF (Id_Style = Uic_S)  AND				{ UIC }
	  Match_Uic (Candidate_Uic,Dancer_Id)  THEN
	    Matched := TRUE
	ELSE
       IF Id_Style = Ident_S  THEN				{ Rights ID }
	  FOR Spot := 1 TO Rightsize  DO
	      IF Dancer_Id.UIC$L_UIC = Candidate_Rights[Spot].L0  THEN
		 Matched := TRUE ;
       END ;

    IF Matched	THEN
       IF Negated  THEN
	  Match_List := CLI$_NEGATED
	ELSE
	  Match_List := CLI$_PRESENT ;
    END ;
  END ; 	{ of Match_List }


PROCEDURE Parse_Number_List (VAR List_Str : VARYING [Len] OF CHAR) ;

  CONST Max_Str_Sz = 1024 ;

  VAR	I , Cycle , Spot , Num , Range_Start : INTEGER := 0 ;
	Range : BOOLEAN := FALSE ;
	Next_Chr : CHAR := Null ;
	Str : VARYING [Max_Str_Sz] OF CHAR := ZERO ;

  BEGIN 	{ Parse_Number_List }
  IF Bugger.D2	THEN
     WRITELN (Dbgfile,'*** Parsing numbers: Source = ',List_Str) ;

  IF (FIND_NONMEMBER (List_Str,Digits+Separators) <> 0) OR (Len = 0)  THEN
       LIB$SIGNAL (Jump__Baddata) ;

  WHILE LENGTH(List_Str) > 0  DO
    BEGIN
    Cycle := Cycle + 1 ;
    IF Bugger.D2  THEN
       WRITELN (Dbgfile,'*** Parsing numbers: -------- CYCLE ',Cycle:1,
			' --------') ;

    Spot := FIND_MEMBER (List_Str,Separators) ;
    IF Bugger.D2  THEN
       WRITELN (Dbgfile,'*** Parsing numbers: Spot  = ',Spot:1) ;

    IF Spot = 0  THEN
       BEGIN
       Str := List_Str ;
       List_Str := '' ;
       END
     ELSE
    IF (Spot = 1) OR (Spot = LENGTH(List_Str))	THEN
       LIB$SIGNAL (Jump__Baddata)	{ Adjacent/hanging separator }
     ELSE
       BEGIN
       Str := SUBSTR (List_Str,1,Spot-1) ;
       Next_Chr := List_Str[Spot] ;
       List_Str := SUBSTR (List_Str,Spot+1,LENGTH(List_Str)-Spot) ;
       IF Next_Chr = '-'  THEN
	  BEGIN
	  IF Range  THEN
	     LIB$SIGNAL (Jump__Baddata) ;	{ Already in a range! }
	  Range := TRUE ;
	  END ;
       END ;

    IF Bugger.D2  THEN
       BEGIN
       WRITELN (Dbgfile,'*** Parsing numbers: Str   = ',Str) ;
       WRITELN (Dbgfile,'*** Parsing numbers: Rem   = ',List_Str) ;
       WRITELN (Dbgfile,'*** Parsing numbers: Range =' ,Range) ;
       WRITELN (Dbgfile,'*** Parsing numbers: Next  = ',Next_Chr) ;
       END ;

    READV (Str,Num) ;
    IF (Num < 1) OR (Num > Max_Scm_Num)  THEN
       LIB$SIGNAL (Jump__Baddata) ;
    IF Default_Scm = 0	THEN
       Default_Scm := Num ;	{ First in list is default cmd if only one }

    IF Bugger.D2  THEN
       WRITELN (Dbgfile,'*** Parsing numbers: Num   = ',Num:1) ;

    IF Range  THEN
       IF Range_Start = 0  THEN
	  Range_Start := Num
	ELSE
	  BEGIN
	  IF Num <= Range_Start  THEN
	     LIB$SIGNAL (Jump__Baddata) ;
	  FOR I := Range_Start TO Num  DO
	      BEGIN
	      Valid_Scm_Numbers[I] := TRUE ;
	      Scm_Count := Scm_Count + 1 ;
	      END ;
	  Range := FALSE ;
	  Range_Start := 0 ;
	  END
     ELSE
       BEGIN
       Valid_Scm_Numbers[Num] := TRUE ;
       Scm_Count := Scm_Count + 1 ;
       END ;
    END ;	{ While }

    IF Bugger.D3  THEN
       FOR I := 1 TO Max_Scm_Num  DO
	   IF Valid_Scm_Numbers[I]	THEN
	      WRITELN (Dbgfile,'*** Parsing numbers: Valid = ',I:4) ;

  END ; 	{ Parse_Number_List }


PROCEDURE Validate_Access ;

{ For the type of user running the program, validate the user's access
  against the target user's UAF record.

  Non-SysProgs can jump using /NOEXACT only to users who do not have any
  privileges other than "minor" privileges, or to those to whom access is
  granted in the Access List file. "Minor" privileges are specified through
  a SITE-specific privilege list using the JUMP_MINOR_PRIVS logical name. }

  VAR	I   : INTEGER  := 0 ;
	Rst : UNSIGNED := 0 ;
        Chekov ,                        { User has JUMP's rights ID? }
	Priv_Target ,			{ Target is "privileged" ? }
	Id_Check_Ok : BOOLEAN := FALSE ;{ JUMP_ACCESS rights ID check ok ? }
	Access : Access_Status := Unspecified ; { Result of access list checks }
	Jump_Id : UIC$TYPE := ZERO ;

  FUNCTION Check_Access_And_Options : Access_Status ;

  { Determine if the caller is specifically authorised to access the target
    user in the access list data file. }

    VAR I , Spot , Skip : INTEGER := 0 ;
	Rst , Target_Status : UNSIGNED := 0 ;
	Done : BOOLEAN := FALSE ;
	Option , Scanning : BOOLEAN := TRUE ;
	Str , Src , Dst , Opt , Line : VARYING [120] OF CHAR := '' ;
	Buf : VARYING [1024] OF CHAR := '' ;
	Result : Access_Status := Unspecified ;
	Access : TEXT ;

    BEGIN	{ Check_Access_And_Options }
    Check_Access_And_Options := Unspecified ;			{ Default }

    High_Security := TRUE ;
    OPEN (Access,FILE_NAME:=Access_List,HISTORY:=READONLY,SHARING:=READONLY,
		 USER_ACTION:=Special_File_Open,Error:=CONTINUE) ;
    Rst := STATUS (Access) ;
    IF Rst = PAS$K_SUCCESS  THEN
       BEGIN
       RESET (Access) ;
       Access_List := Open_File_Name ;

       WHILE NOT (EOF (Access) OR Done)  DO
	 BEGIN
	 READLN (Access,Line) ;
	 Str_Compress (Line,Line,TRUE) ;		{ Squeeeeeeze! }
	 Rst := STR$UPCASE (%DESCR Line,Line) ;
	 Zip (Rst) ;
	 IF (Line.LENGTH > 0) AND_THEN (Line[1] <> '#')  THEN  { Not a comment }
	    BEGIN
	    Spot := INDEX (Line,'#') ;			{ Trailing comment? }
	    IF Spot > 0  THEN
	       Line := SUBSTR (Line,1,Spot-1) ;
	    Spot := INDEX (Line,'\') ;
	    IF Spot > 0  THEN				{ Continued ... ? }
	       Buf := Buf + SUBSTR (Line,1,Spot-1)
	     ELSE
	    IF Line = '!!!'  THEN			{ Terminate processing }
	       Done := TRUE
	     ELSE
	       BEGIN
	       Buf  := Buf + Line ;
	       Spot := INDEX (Buf,':') ;
	       IF Spot = 0  THEN  $EXIT (Jump__Baddata) ;
	       Src  := SUBSTR (Buf,1,Spot-1) ;
	       Dst  := SUBSTR (Buf,Spot+1,Buf.LENGTH-Spot) ;
	       Spot := INDEX (Dst,':') ;
	       IF Spot = 0  THEN
		  Opt := ''
		ELSE
		  BEGIN
		  Opt := SUBSTR (Dst,Spot+1,Dst.LENGTH-Spot) ;
		  Dst := SUBSTR (Dst,1,Spot-1) ;
		  END ;
	       IF (Src = '') OR (Dst = '')  THEN
		  $EXIT (Jump__Baddata) ;

	       { Check if target user is in valid target list. }

	       Target_Status := Match_List (New_User,Dst,New_Uic,Target_Rights);

	       { If caller is in valid caller list, determine access status. }

	       IF Match_List
		    (Orig_User,Src,Orig_Uic,Caller_Rights) = CLI$_PRESENT  THEN
		  BEGIN
		  IF Target_Status = CLI$_PRESENT  THEN  { Newuser in target...}
		     Result := Granted			 { ... NOT negated }
		   ELSE
		  IF Target_Status = CLI$_NEGATED  THEN  { ... negated! }
		     Result := Denied
		   ELSE
		     Result := Unspecified ;		 { Not there }

		  IF Result <> Unspecified  THEN
		     BEGIN
		     Done := TRUE ;
		     Check_Access_And_Options := Result ;
		     END ;
		  END ;

	       { If access is Granted, check and action any options ... }

	       IF (Result = Granted) AND (Opt <> '')  THEN
		  FOR I := 1 TO Opt.LENGTH  DO
		    IF Scanning  THEN
		      BEGIN
		      CASE Opt[I]  OF
			'!': Option := FALSE ;			{ Negate }
			'A': Notify.After  := Option ;
			'B': Notify.Before := Option ;
			'C': IF NOT Option  THEN	 { Not negatable }
				$EXIT (Jump__Baddata)
			      ELSE
				Chain := FALSE ;
			'E': BEGIN
			     IF (Option AND NOT Real_Mccoy) OR
				(Real_Mccoy AND NOT Option)  THEN
				  Check_Access_And_Options := Denied ;
			     IF (NOT Option) AND (Opt <> '!E')	THEN	{ ONLY }
				$EXIT (Jump__Baddata) ;
			     END ;
			'I': Notify.Include_Log := Option AND
						   Record_Session AND
						   Notify.By_Mail  ;
			'M': Notify.By_Mail := Option ;
			'N': BEGIN
			     Notify.All_Bits := Option::Unsigned8 * 16#Ffffffff;
			     Notify.Include_Log := Record_Session AND Option ;
			     END ;
			'O': Notify.By_Opcom := Option ;
			'R': IF Opt[1] = 'S'  THEN
				$EXIT (Jump__Baddata)
			      ELSE
			     IF Secure_Mode AND NOT Option  THEN
				$EXIT (Jump__Conflict)
			      ELSE
				Record_Session := Option ;
			'U': BEGIN
			     IF NOT Option  THEN
				$EXIT (Jump__Baddata) ;
			     IF (Opt.LENGTH > I) AND_THEN
				(Opt[I+1] IN Digits)  THEN
				BEGIN
				Str := SUBSTR (Opt,I+1,Opt.LENGTH-I) ;
				Spot := FIND_NONMEMBER (Str,Digits+Separators) ;
				IF Spot > 0  THEN
				   BEGIN
				   Str := SUBSTR (Str,1,Spot-1) ;
				   Skip := Spot - 1 ;
				   END
				 ELSE
				   Skip := Str.LENGTH ;
				Parse_Number_List (Str) ;
				IF Scm_Selection > 0  THEN
				   IF NOT Valid_Scm_Numbers[Scm_Selection]  THEN
				      LIB$SIGNAL (Jump__Unkscmnum)
				    ELSE
				      Spot := Scm_Selection
				 ELSE
				   Spot := Default_Scm ;
				WRITEV (Str,Spot:1) ;
				Get_Logical_Name (
					Lognam	:= 'JUMP_SINGLE_CMD_' + Str,
					Default := 'MYDOGHASFLEAS',
					Table	:= 'JUMP_SINGLE_COMMANDS',
					Actual	:= Str) ;
				IF Single_Cmd_Mode  THEN	{ Imposter?! }
				   LIB$SIGNAL (Jump__Noovrsnglcmd) ;
				IF Str = 'MYDOGHASFLEAS'  THEN
				   LIB$SIGNAL  (Jump__Nolnmforscm) ;
				Single_Cmd := Str ;
				Single_Cmd_Mode := TRUE ;
				IF (Scm_Selection = 0) AND (Scm_Count > 1)  THEN
				   LIB$SIGNAL (Jump__Scmcmddef) ;  { Warn them }
				END
			      ELSE
			     IF Scm_Selection > 0  THEN
				LIB$SIGNAL (Jump__Scmnotsel)
			      ELSE
			     IF NOT Single_Cmd_Mode  THEN	{ Mandatory! }
				$EXIT (Jump__Scmreqcmd) ;
			     END ;
			'S': BEGIN
			     IF I <> 1	THEN
				$EXIT (Jump__Baddata) ;
			     Secure_Mode := TRUE ;
			     Notify.All_Bits := 16#Ffffffff ;
			     Record_Session  := TRUE ;
			     END ;
			'X': BEGIN
			     IF Opt <> 'X'  THEN	{ Must be just 'X' }
				$EXIT (Jump__Baddata) ;
			     Houdini := TRUE ;
			     END ;
			'+',
			'=': BEGIN			{ Buf no longer in use }
			     IF NOT Option  THEN
				$EXIT (Jump__Baddata) ;
			     Buf := SUBSTR (Opt,I+1,Opt.LENGTH-I) ;
			     IF Opt[I] = '+'  THEN
				Notify_Maillist := Notify_Maillist + Buf
			      ELSE
				Notify_Maillist := Buf ;
			     Scanning := FALSE ;	{ Stop here }
			     END ;
			OTHERWISE
			     IF Skip > 0  THEN
				Skip := Skip - 1
			      ELSE
				LIB$SIGNAL (Jump__Baddata) ;
			END ;	{ of Case }

		      IF Opt[I] <> '!'	THEN		{ Reset }
			 Option := TRUE ;
		      END ;	{ of FOR }

	       Buf := '' ;		{ Logical end of line }
	       END ;	{ of continue processing (not "!!!") }
	    END ;	{ of not a comment }
	 END ;		{ of WHILE }

       IF Bugger.Info  THEN
	  BEGIN
	  WRITELN (Dbgfile,'*** Target access:',Result) ;
	  IF Result <> Unspecified  THEN
	     WRITELN (Dbgfile,'*** Options: ',Opt) ;
	  END ;

       IF Buf <> ''  THEN		{ Continuation?  Huh? }
	  $EXIT (Jump__Baddata) ;
       CLOSE (Access) ;
       END		{ of successful file open }
     ELSE
    IF Rst <> PAS$K_FILNOTFOU  THEN	{ Ignore FILNOTFOU - file is optional }
       BEGIN
       LIB$SIGNAL (Jump__Badaccfil) ;
       LIB$STOP (Condition_Code(Rst)) ;
       END ;

    IF Bugger.Info AND (Rst = PAS$K_FILNOTFOU)	THEN
       WRITELN (Dbgfile,'*** Access List file does not exist') ;

    IF (Notify.Before OR Notify.After) AND
       NOT (Notify.By_Mail OR Notify.By_Opcom)	THEN
	 $EXIT (Jump__Badnotify) ;

    IF Notify.Include_Log AND
       NOT (Record_Session AND Notify.By_Mail)	THEN
	 $EXIT (Jump__Badinclude) ;

    IF System_Secure_Mode AND NOT Houdini  THEN
       BEGIN
       Notify.All_Bits := 16#Ffffffff ;
       Record_Session  := TRUE ;
       Secure_Mode := TRUE ;
       END ;

    IF (Notify.By_Mail OR Notify.By_Opcom) AND
       NOT (Notify.Before OR Notify.After)  THEN
       BEGIN
       Notify.After  := TRUE ;
       Notify.Before := TRUE ;
       LIB$SIGNAL (Jump__Fixnotify) ;
       END ;

    IF Secure_Mode  THEN
       BEGIN
       IF Notify.By_Mail  THEN
	  BEGIN
	  Notify.Vamoose := TRUE ;
	  Notify.Before  := TRUE ;	{ After is too late }
	  END ;
       IF Escape_Hatch	THEN
	  BEGIN
	  Escape_Hatch := FALSE ;
	  LIB$SIGNAL (Jump__Noescape) ;
	  END ;
       END ;

    IF (Scm_Selection > 0) AND NOT Single_Cmd_Mode  THEN
       LIB$SIGNAL (Jump__Selnotreq) ;

    Suspect := Suspect AND (Secure_Mode OR Record_Session) ;
    END ;	{ of Check_Access_And_Options }


  PROCEDURE Validate_Maillist (Listname : VARYING [Sz] OF CHAR) ;

  { Check the addresses in the Mailing List file to see if there are any
    dodgy logical names (not defined /SYSTEM /EXEC).  The validity of
    mail addresses is not tested. }

    VAR  Rst , Spot : INTEGER := 0 ;
	 Done : BOOLEAN := FALSE ;
	 Dest_Addr , Result : VARYING [255] OF CHAR := '' ;
	 Dis_File : TEXT ;

    BEGIN	{ Validate_Maillist }
    Dest_Addr := SUBSTR (Listname,2,Sz-1) ;  { Remove "@" }

    High_Security := TRUE ;
    OPEN (Dis_File,FILE_NAME:=Dest_Addr,HISTORY:=READONLY,SHARING:=READONLY,
		   USER_ACTION:=Special_File_Open,Error:=CONTINUE) ;
    Rst := STATUS (Dis_File) ;

    IF Rst <> PAS$K_SUCCESS  THEN
       LIB$SIGNAL (Condition_Code (Rst))   { Map to a condition code value }
     ELSE
       BEGIN
       RESET (Dis_File) ;
       WHILE NOT (Done OR EOF (Dis_File))  DO
	 BEGIN
	 READLN (Dis_File,Dest_Addr) ;
	 Str_Compress (Dest_Addr,Dest_Addr,TRUE) ;
	 Spot := INDEX (Dest_Addr,'!') ;
	 IF Spot > 1  THEN
	    Dest_Addr := SUBSTR (Dest_Addr,1,Spot-1) ;
	 IF Spot <> 1  THEN
	    BEGIN
	    Rst := Get_Logical_Name (Dest_Addr,'###',Result,
				     'LNM$FILE_DEV',PSL$C_USER) ;
	    IF NOT ((Rst = SS$_NOLOGNAM) OR Secure_Logical)  THEN
	       LIB$SIGNAL (Jump__Invlnm,1,
			   %STDESCR SUBSTR (Dest_Addr,1,Dest_Addr.LENGTH)) ;
	    Done := NOT ODD (Rst) ;
	    END ;
	 END ;
       CLOSE (Dis_File) ;
       END ;
    END ;	{ of Validate_Maillist }


  PROCEDURE Validate_Notify_List ;

  { Check the addresses in the JUMP_NOTIFY_MAILLIST logical name to see if
    there are any dodgy logical names (not defined /SYSTEM /EXEC).  The
    validity of mail addresses is not tested. }

    VAR  Rst , I : INTEGER := 0 ;
	 Dest_Addr ,
	 Result ,
	 Addr_List : VARYING [255] OF CHAR := '' ;

    BEGIN	{ Validate_Notify_List }

    Addr_List := Notify_Maillist ;

    WHILE Addr_List.LENGTH > 0	DO
      BEGIN
      I := INDEX (Addr_List,',') ;
      IF I > 0	THEN
	 BEGIN
	 Dest_Addr := SUBSTR (Addr_List,1,I-1) ;
	 Addr_List := SUBSTR (Addr_List,I+1,Addr_List.LENGTH-I) ;
	 END
       ELSE
	 BEGIN
	 Dest_Addr := Addr_List ;
	 Addr_List := '' ;
	 END ;

      IF Dest_Addr.LENGTH > 0  THEN
	 BEGIN
	 IF Dest_Addr[1] = '@'	THEN
	    Validate_Maillist (Dest_Addr)
	  ELSE
	    BEGIN
	    Rst := Get_Logical_Name (Dest_Addr,'###',Result,
				     'LNM$FILE_DEV',PSL$C_USER) ;
	    IF NOT ((Rst = SS$_NOLOGNAM) OR Secure_Logical)  THEN
	       LIB$SIGNAL (Jump__Invlnm,1,
			   %STDESCR SUBSTR (Dest_Addr,1,Dest_Addr.LENGTH)) ;
	    END ;
	 END ;
      END ;
  END ; 	{ of Validate_Notify_List }


  BEGIN 	{ Validate_Access }

  { Check that the invoker has the required access to run this program.
    This is independant of any installed privileges.  Identify SysProgs.
    If need be, check to see if process has JUMP_ACCESS rights ID. }

  IF Double_Check  THEN
     BEGIN
     IF NOT Parse_Ident ('JUMP_ACCESS',Jump_Id,Rst,1)  THEN	{ Rights ID }
	$EXIT (Jump__Ivident) ;

     I := 1 ;
     REPEAT
       Chekov := Jump_Id.UIC$L_UIC = Caller_Rights[I].L0 ;
       I := I + 1 ;
     UNTIL Chekov OR (I > Rightsize) ;

     IF Bugger.Info  THEN
	BEGIN
	WRITE (Dbgfile,'*** Double Check JUMP_ACCESS: ') ;
	IF Chekov  THEN
	   WRITELN (Dbgfile,'Granted!')
	 ELSE
	   WRITELN (Dbgfile,'NOT Granted!')
	END ;
     END ;

  Id_Check_Ok := (Double_Check AND Chekov) OR (NOT Double_Check) ;

  IF NOT Real_Mccoy  THEN               { For "poor man's" JUMP only. }
     BEGIN
     IF Proc_Cnt > 0  THEN              { Don't jump if we have subprocesses }
	$EXIT (Jump__Nosub) ;

     IF Pid <> Master_Pid  THEN         { Don't jump if we are a subprocess }
	$EXIT (Jump__Noinsub) ;
     END ;

  IF Figment AND NOT Sysprog  THEN
     BEGIN
     Audit_Jump (FALSE) ;
     $EXIT (Jump__Nopriv) ;
     END ;

  IF NOT Get_Target_Info  THEN
     IF Figment AND Alter_Ego AND NOT (Transmute OR Real_Mccoy)  THEN
	LIB$SIGNAL (Jump__Invuser)
      ELSE
	LIB$STOP (Jump__Invuser) ;	{ Fatal }

  Access := Check_Access_And_Options ;

  Priv_Target := ((Auth_Priv - Minor_Privs) <> []) OR
		 ((Def_Priv  - Minor_Privs) <> []) OR
		 (New_Uic.UIC$V_GROUP <= Max_Sys_Group) ;

  IF NOT (Sysprog OR
		(
		 Id_Check_Ok AND
		 Auditing AND
		 NOT Alter_Ego AND
		   (
		     (Access = Granted) OR
		       (
			 (
			  (Operator AND NOT Priv_Target) OR
			  (Narcissus AND (New_User = Orig_User))
			 ) AND
			 (Access <> Denied)
		       )
		   )
		)
	 )  THEN
     BEGIN
     Audit_Jump (FALSE) ;
     $EXIT (Jump__Nopriv) ;
     END ;

  IF Alter_Ego AND (New_User = Orig_User)  THEN 	{ Change username? }
     BEGIN
     LIB$SIGNAL (Jump__Sameuser,1,
		 %STDESCR SUBSTR (Orig_User,1,Orig_User.LENGTH)) ;
     $EXIT ;
     END ;

  IF Transmute AND
     (New_Uic.UIC$L_UIC = Orig_Uic.UIC$L_UIC)  THEN { Change UIC etc? }
     BEGIN
     LIB$SIGNAL (Jump__Sameuic,1,%IMMED Orig_Uic.UIC$L_UIC) ;
     $EXIT ;
     END ;

  IF Sysprog AND (Access = Denied)  THEN
     LIB$SIGNAL (Jump__Denied) ;

  IF (Target_Flags.UAI$V_RESTRICTED OR Target_Flags.UAI$V_CAPTIVE)  THEN
     BEGIN
     IF Sysprog OR (Access = Granted)  THEN
	LIB$SIGNAL (Jump__Restrict)
      ELSE
	LIB$STOP (Jump__Restrict) ;		{ Fatal }

     IF Escape_Hatch  THEN                      { Don't allow escape }
	BEGIN
	Escape_Hatch := FALSE ;
	LIB$SIGNAL (Jump__Noescape) ;
	END ;
     END ;

  IF Target_Flags.UAI$V_DISACNT  THEN
     IF Figment  THEN
	LIB$SIGNAL (Jump__Disabled)
      ELSE
	LIB$STOP (Jump__Disabled) ;		{ Fatal }

  IF Secure_Mode AND Notify.By_Mail  THEN  Validate_Notify_List ;
  END ; 	{ of Validate_Access }


[ASYNCHRONOUS] FUNCTION Mail_Error_Handler (
			VAR Sa : Sig_Args;
			VAR Ma : Mech_Args) : [UNSAFE] INTEGER ;

{ A simple handler that always allows MAIL to handle the condition ...
  but ensures the status returned is accurate!	This is required for the
  MAIL$SEND_MESSAGE routine which fails to return its status to the
  caller! 8-(( }

  BEGIN
  Mail_Error_Status  := Sa[1] ; 	{ Primary condition value }
  Mail_Error_Handler := SS$_RESIGNAL ;	{ Propagate back to MAIL }
  END ;


FUNCTION Send_Mail_Message (To_Address : PACKED ARRAY [L0..H0:INTEGER] OF CHAR ;
			    Subj_Line  : PACKED ARRAY [L1..H1:INTEGER] OF CHAR ;
			    Msg_Text   : PACKED ARRAY [L2..H2:INTEGER] OF CHAR
		) : INTEGER ;

{ Send a mail message using Callable MAIL.  The Msg_Text is treated firstly
  as the file specification of a message file to be used in a call to
  MAIL$SEND_ADD_BODYPART.  If it is not a valid specification of an existing
  file, the Msg_Text will be inserted into the mail message verbatim. }

  VAR  I , Rst	 : INTEGER  := 0 ;
       Mail_Ctx  ,
       Context	 : [VOLATILE] UNSIGNED := 0 ;
       Mail_List : Item_List_Template(2) := ZERO ;
       Addr_List,
       Dest_Addr : STRING (MAX(510,SIZE(To_Address))) := ZERO ;
       Msg_Buff,
       Msg_Line  : STRING (SIZE(Msg_Text)) := ZERO ;
       Msg_File  : VARYING [255] OF CHAR := '' ;
       Dis_File  : TEXT ;

  FUNCTION Distribution_List : INTEGER ;

    VAR  Rst , Spot : INTEGER := 0 ;
	 Done : BOOLEAN := FALSE ;

    BEGIN	{ Distribution_List }
    Dest_Addr := SUBSTR (Dest_Addr,2,LENGTH(Dest_Addr)-1) ;  { Remove "@" }

    High_Security := TRUE ;
    OPEN (Dis_File,FILE_NAME:=Dest_Addr,HISTORY:=READONLY,SHARING:=READONLY,
		   USER_ACTION:=Special_File_Open,Error:=CONTINUE) ;
    Rst := STATUS (Dis_File) ;

    IF Rst <> PAS$K_SUCCESS  THEN
       Rst := Condition_Code (Rst)	{ Map to a condition code value }
     ELSE
       BEGIN
       RESET (Dis_File) ;
       WHILE NOT (Done OR EOF (Dis_File))  DO
	 BEGIN
	 READLN (Dis_File,Dest_Addr) ;
	 Str_Compress (Dest_Addr,Dest_Addr,TRUE) ;
	 Spot := INDEX (Dest_Addr,'!') ;
	 IF Spot > 1  THEN
	    Dest_Addr := SUBSTR (Dest_Addr,1,Spot-1) ;
	 IF Spot <> 1  THEN
	    BEGIN
	    Mail_List[1].Buffer_Length := LENGTH (Dest_Addr) ;
	    Rst := MAIL$SEND_ADD_ADDRESS (Mail_Ctx,Mail_List,Null_List) ;
	    Done := NOT ODD (Rst) ;
	    END ;
	 END ;
       CLOSE (Dis_File) ;
       END ;
    Distribution_List := Rst ;
    END ;	{ of Distribution_List }

  BEGIN 	{ Send_Mail_Message }

  { Prepare the Mail SEND context }

  Rst := MAIL$SEND_BEGIN (Mail_Ctx,Null_List,Null_List) ;
  IF Notify.Vamoose AND NOT ODD (Rst)  THEN
     LIB$SIGNAL (Jump__Mailfail) ;

  { Set up the To: address(es) }

  Mail_List[1].Buffer_Length := 0 ;			{ Initially ... }
  Mail_List[1].Item_Code     := MAIL$_SEND_USERNAME ;
  Mail_List[1].Buffer_Addr   := IADDRESS (Dest_Addr.BODY) ;
  Mail_List[1].Return_Addr   := 0 ;

  Mail_List[2].Terminator    := 0 ;	{ Terminate the item list }

  Addr_List := To_Address ;

  WHILE LENGTH (Addr_List) > 0	DO
    BEGIN
    I := INDEX (Addr_List,',') ;
    IF I > 0  THEN
       BEGIN
       Dest_Addr := SUBSTR (Addr_List,1,I-1) ;
       Addr_List := SUBSTR (Addr_List,I+1,LENGTH(Addr_List)-I) ;
       END
     ELSE
       BEGIN
       Dest_Addr := Addr_List ;
       Addr_List := '' ;
       END ;
    IF LENGTH (Dest_Addr) > 0  THEN
       BEGIN
       IF Dest_Addr[1] = '@'  THEN
	  Rst := Distribution_List
	ELSE
	  BEGIN
	  Mail_List[1].Buffer_Length := LENGTH (Dest_Addr) ;
	  Rst := MAIL$SEND_ADD_ADDRESS (Mail_Ctx,Mail_List,Null_List) ;
	  END ;
       IF Notify.Vamoose AND NOT ODD (Rst)  THEN
	  LIB$SIGNAL (Jump__Mailfail) ;
       END ;
    END ;

  { Set up the Subject line }

  Mail_List[1].Buffer_Length := LENGTH (Subj_Line) ;
  Mail_List[1].Item_Code     := MAIL$_SEND_SUBJECT ;
  Mail_List[1].Buffer_Addr   := IADDRESS (Subj_Line) ;
  Mail_List[1].Return_Addr   := 0 ;

  Mail_List[2].Terminator    := 0 ;

  Rst := MAIL$SEND_ADD_ATTRIBUTE (Mail_Ctx,Mail_List,Null_List) ;
  IF Notify.Vamoose AND NOT ODD (Rst)  THEN
     LIB$SIGNAL (Jump__Mailfail) ;

  { Determine if Msg_Text is a valid specification of an existing file. }

  Rst := LIB$FIND_FILE (Filespec := Msg_Text,
			Resultant_Filespec := %DESCR Msg_File,
			Context  := Context,
			Flags	 := 0) ;	{ No wildcards }

  IF ODD (Rst)	THEN	{ File exists - attach it. }
     BEGIN
     Mail_List[1].Buffer_Length := LENGTH (Msg_Text) ;
     Mail_List[1].Item_Code	:= MAIL$_SEND_FILENAME ;
     Mail_List[1].Buffer_Addr	:= IADDRESS (Msg_Text) ;
     Mail_List[1].Return_Addr	:= 0 ;

     Mail_List[2].Terminator	:= 0 ;

     Rst := MAIL$SEND_ADD_BODYPART (Mail_Ctx,Mail_List,Null_List) ;
     IF Notify.Vamoose AND NOT ODD (Rst)  THEN
	LIB$SIGNAL (Jump__Mailfail) ;
     END
   ELSE 	{ Split Msg_Text at LF characters and send record by record }
     BEGIN
     Mail_List[1].Item_Code   := MAIL$_SEND_RECORD ;
     Mail_List[1].Buffer_Addr := IADDRESS (Msg_Line.BODY) ;
     Msg_Buff := Msg_Text ;
     WHILE LENGTH (Msg_Buff) > 0  DO
       BEGIN
       I := INDEX (Msg_Buff,Lf) ;
       IF I > 0 THEN
	  BEGIN
	  Msg_Line := SUBSTR (Msg_Buff,1,I-1) ;
	  Msg_Buff := SUBSTR (Msg_Buff,I+1,LENGTH(Msg_Buff)-I) ;
	  END
	ELSE
	  BEGIN
	  Msg_Line := Msg_Buff ;
	  Msg_Buff := '' ;
	  END ;
       Mail_List[1].Buffer_Length := LENGTH (Msg_Line) ;
       Rst := MAIL$SEND_ADD_BODYPART (Mail_Ctx,Mail_List,Null_List) ;
       IF Notify.Vamoose AND NOT ODD (Rst)  THEN
	  LIB$SIGNAL (Jump__Mailfail) ;
       END ;
     END ;

  { The message is complete.  Let's send it.  Mail_Error_Status is
    updated by the error handler, so we can handle errors correctly
    when an exception occurs. Also, set a default success status for
    Mail_Error_Status so that a successful call is not flagged as a
    false error because the exception handler doesn't get called. }

  Mail_Error_Status := SS$_NORMAL ;		{ Default value }
  ESTABLISH (Mail_Error_Handler) ;
  Rst := MAIL$SEND_MESSAGE (Mail_Ctx,Null_List,Null_List) ;
  REVERT ;

  IF Notify.Vamoose AND
     ((NOT ODD (Rst)) OR
      (NOT ODD (Mail_Error_Status)))  THEN
     LIB$SIGNAL (Jump__Mailfail) ;
  Send_Mail_Message := Rst ;

  Rst := MAIL$SEND_END (Mail_Ctx,Null_List,Null_List) ;
  IF Notify.Vamoose AND NOT ODD (Rst)  THEN
     LIB$SIGNAL (Jump__Mailfail) ;
  END ; 	{ of Send_Mail_Message }


PROCEDURE Update_Uaf ;

{ Set the Last Interactive Login time, and reset the Login Failures to
  zero. }

  VAR	Fails : $UWORD := 0 ;
	Rst : UNSIGNED := 0 ;
	Now : $UQUAD := ZERO ;
	Item_List : Item_List_Template (3) := ZERO ;

  BEGIN 	{ Update_UAF }
  GETTIMESTAMP (Time_Now) ;
  Now := Time_Now.Binary_Time ;

  Item_List[1].Buffer_Length := 2 ;
  Item_List[1].Item_Code     := UAI$_LOGFAILS ;
  Item_List[1].Buffer_Addr   := IADDRESS (Fails) ;
  Item_List[1].Return_Addr   := 0 ;

  Item_List[2].Buffer_Length := 8 ;
  Item_List[2].Item_Code     := UAI$_LASTLOGIN_I ;
  Item_List[2].Buffer_Addr   := IADDRESS (Now) ;
  Item_List[2].Return_Addr   := 0 ;

  Item_List[3].Terminator    := 0 ;	{ Terminate the item list }

  Rst := $SETUAI (Usrnam := %STDESCR SUBSTR (New_User,1,New_User.LENGTH),
		  Itmlst := Item_List) ;

  IF NOT ODD (Rst) AND_THEN (Rst <> RMS$_RNF)  THEN
     LIB$SIGNAL (Rst) ;
  END ; 	{ of Update_UAF }


PROCEDURE Getmem (VAR Location ,
		  Pointer  : [UNSAFE] Unsigned_Ptr) ; EXTERNAL ;

PROCEDURE Putmem (VAR Location ,
		  Pointer  : [UNSAFE] Unsigned_Ptr) ; EXTERNAL ;

[ASYNCHRONOUS,Check(None)] PROCEDURE Getuser ; EXTERNAL ;

[ASYNCHRONOUS,Check(None)] PROCEDURE Setuser ; EXTERNAL ;

PROCEDURE Change_Persona (New_Username : VARYING [Len] OF CHAR) ;
			    EXTERNAL ;

PROCEDURE Change_Username (New_Username : VARYING [Len] OF CHAR ;
			   Faking : BOOLEAN := FALSE) ;
			     EXTERNAL ;

PROCEDURE Change_Uic (New_Uic : UIC$TYPE ;
		      Faking : BOOLEAN := FALSE) ;
			EXTERNAL ;

PROCEDURE Kangaroo ;

{ Do all that is required to JUMP /NOEXACT to the new user, or to return
  to the original user.  This is the "poor man's" jump. }

  VAR	Rst , Attributes : UNSIGNED := 0 ;
	Aclstr : VARYING [64] OF CHAR := '' ;
	Aclent : PACKED ARRAY [1..32] OF CHAR := '' ;
	Grptbl : PACKED ARRAY [1..16] OF CHAR := '' ;
	Item_List : Item_List_Template (3) := ZERO ;

  BEGIN 	{ Kangaroo }

  Change_Uic (New_Uic) ;	{ *** Set new UIC *** }

  { Set new default directory. }

  Set_Default_Dir (Def_Dir) ;

  { Set new default disk. }

  Item_List[1].Buffer_Length := Def_Dev.LENGTH ;
  Item_List[1].Item_Code     := LNM$_STRING ;
  Item_List[1].Buffer_Addr   := IADDRESS (Def_Dev.BODY) ;
  Item_List[1].Return_Addr   := 0 ;

  Item_List[2].Terminator    := 0 ;

  Rst := $CRELNM (Tabnam := %STDESCR 'LNM$PROCESS',
		  Lognam := %STDESCR 'SYS$DISK',
		  Acmode := PSL$C_SUPER,
		  Itmlst := Item_List) ;
  Zip (Rst) ;

  { Point LNM$GROUP logical to group table for new UIC. }

  Attributes := LNM$M_TERMINAL ;
  Grptbl := 'LNM$GROUP_' + OCT (New_Uic.UIC$V_GROUP,6,6) ;

  Item_List[1].Buffer_Length := 4 ;
  Item_List[1].Item_Code     := LNM$_ATTRIBUTES ;
  Item_List[1].Buffer_Addr   := IADDRESS (Attributes) ;
  Item_List[1].Return_Addr   := 0 ;

  Item_List[2].Buffer_Length := SIZE (Grptbl) ;
  Item_List[2].Item_Code     := LNM$_STRING ;
  Item_List[2].Buffer_Addr   := IADDRESS (Grptbl) ;
  Item_List[2].Return_Addr   := 0 ;

  Item_List[3].Terminator    := 0 ;

  Rst := $CRELNM (Tabnam := %STDESCR 'LNM$PROCESS_DIRECTORY',
		  Lognam := %STDESCR 'LNM$GROUP',
		  Acmode := PSL$C_KERNEL,
		  Itmlst := Item_List) ;
  Zip (Rst) ;

  { If going to a different UIC, allow the current LNM$JOB logical name table
    to be accessed by the new UIC.  If returning to original user, remove the
    ACL access previously applied. }

  Item_List[2].Terminator := 0 ;

  IF New_User = Orig_User  THEN 	{ Return to original user. }
     BEGIN
     Item_List[1].Buffer_Length := 0 ;
     Item_List[1].Item_Code	:= ACL$C_DELETEACL ;
     Item_List[1].Buffer_Addr	:= 0 ;
     Item_List[1].Return_Addr	:= 0 ;
     END
   ELSE 				{ Allow access to new user. }
     BEGIN
     Format_User (New_Uic) ;
     Aclstr := '(IDENTIFIER=' + Uic_Str + ',ACCESS=READ+WRITE)' ;
     Rst := $PARSE_ACL (Aclstr := SUBSTR (Aclstr.BODY,1,Aclstr.LENGTH),
			Aclent := %STDESCR Aclent) ;
     Zip (Rst) ;

     Item_List[1].Buffer_Length := INT (Aclent[1]) ;
     Item_List[1].Item_Code	:= ACL$C_ADDACLENT ;
     Item_List[1].Buffer_Addr	:= IADDRESS (Aclent) ;
     Item_List[1].Return_Addr	:= 0 ;
     END ;

  Rst := $CHANGE_ACL (Objtyp := ACL$C_LOGICAL_NAME_TABLE,
		      Objnam := 'LNM$JOB',
		      Itmlst := Item_List) ;
  Zip (Rst) ;
  END ; 	{ of Kangaroo }


PROCEDURE Display_Jump ;

{ Display data about the requested non-exact jump. }

  BEGIN 	{ Display_Jump }
  IF Transmute	THEN
     BEGIN
     LIB$SIGNAL (Jump__Jumped,4,
		 %STDESCR SUBSTR (Orig_User,1,Orig_User.LENGTH),
		 %STDESCR SUBSTR (New_User,1,New_User.LENGTH),
		 %STDESCR SUBSTR (Uic_Str,1,Uic_Str.LENGTH),
		 %STDESCR SUBSTR (Eq_Id_Str,1,Eq_Id_Str.LENGTH)) ;
     LIB$SIGNAL (Jump__Default,2,
		 %STDESCR SUBSTR(Def_Dev,1,Def_Dev.LENGTH),
		 %STDESCR SUBSTR(Def_Dir,1,Def_Dir.LENGTH)) ;
     END ;

  IF Alter_Ego	THEN
     LIB$SIGNAL (Jump__Setuser,2,
		 %STDESCR SUBSTR (Orig_User,1,Orig_User.LENGTH),
		 %STDESCR SUBSTR (New_User,1,New_User.LENGTH)) ;
  END ; 	{ of Display_Jump }


PROCEDURE Get_Channel (Device : VARYING [L1] OF CHAR ;
		       VAR Channel : [VOLATILE] $UWORD) ;

{ Assign a channel to a device. }

  VAR Rst : UNSIGNED := 0 ;

  BEGIN 	{ Get_Channel }
  Rst := $ASSIGN (Devnam := Device,
		  Chan	 := Channel) ;
  Zip (Rst) ;
  END ; 	{ of Get_Channel }


PROCEDURE Exit_Handler (Condition : UNSIGNED) ;

{ For an EXACT JUMP only, clean up after the pseudo-terminal session ends or
  on abnormal termination. }

  VAR Rst  : UNSIGNED := 0 ;
      Iosb : Status_Block_Type := ZERO ;

  BEGIN 	{ Exit_Handler }

  $SETAST (0) ; 			{ Disable ASTs }

  { If recording session for an EXACT jump, be sure the log file is closed. }

  IF Record_Session  THEN
     CLOSE (Logfile,Error:=CONTINUE) ;

  { Restore username and UIC if required. }

  IF Use_Persona  THEN
     Change_Persona (Orig_User)
   ELSE
     BEGIN
     IF New_User <> Orig_User THEN
	BEGIN
	New_User.BODY := PAD ('',' ',Max_Username_Len) ;     { Totally blat it }
	New_User := Orig_User ;
	Change_Username (New_User) ;	{ *** Change to original username *** }
	END ;

     IF New_Uic.UIC$L_UIC <> Orig_Uic.UIC$L_UIC THEN
	BEGIN
	New_Uic := Orig_Uic ;
	Change_Uic (New_Uic) ;		{ *** Change to original UIC *** }
	END ;
     END ;

  IF Pchan_Created  THEN
     IF Pseudo_Ft  THEN
	BEGIN
	Rst := PTD$CANCEL (Pchan) ;	{ Cancel I/O to pseudo-terminal }
	Zip (Rst) ;

	Rst := PTD$DELETE (Pchan) ;	{ Delete pseudo-terminal }
	Zip (Rst) ;
	END
      ELSE
	$DASSGN (Pchan) ;		{ Deassign pseudo-terminal channel }

  $CANCEL (Rchan) ;			{ Cancel I/Os on real terminal }

  Rst := $QIOW (Chan := Rchan,		{ Restore original characteristics }
		Func := IO$_SETMODE,
		Iosb := Iosb,
		P1   := Rchars,
		P2   := 12) ;
  Zip (Rst,Iosb[1]) ;

  $DASSGN (Rchan) ;			{ Shut up shop. }

  { Clean up process attributes (if any) ... }

  Rst := $DELLNM (Tabnam := 'LNM$SYSTEM',
		  Lognam := 'JUMP_' + HEX (Pseudo_Pid,8,8),
		  Acmode := PSL$C_EXEC) ;
  IF NOT (ODD (Rst) OR (Rst = SS$_NOLOGNAM))  THEN
     LIB$SIGNAL (Rst) ;

  END ; 	{ of Exit_Handler }


[ASYNCHRONOUS] PROCEDURE Get_Proc_Info (Target_Pid : [VOLATILE] UNSIGNED ;
					VAR Prccnt ,
					    Jobprccnt : [VOLATILE] UNSIGNED) ;

{ Get some process information for the specified PID. }

  VAR	Rst : INTEGER := 0 ;
	Iosb : Status_Block_Type := ZERO ;
	Item_List : Item_List_Template (3) := ZERO ;

  BEGIN 	{ Get_Proc_Info }
  Item_List[1].Buffer_Length := 4 ;
  Item_List[1].Item_Code     := JPI$_PRCCNT ;
  Item_List[1].Buffer_Addr   := IADDRESS (Prccnt) ;
  Item_List[1].Return_Addr   := 0 ;

  Item_List[2].Buffer_Length := 4 ;
  Item_List[2].Item_Code     := JPI$_JOBPRCCNT ;
  Item_List[2].Buffer_Addr   := IADDRESS (Jobprccnt) ;
  Item_List[2].Return_Addr   := 0 ;

  Item_List[3].Terminator    := 0 ;	{ Terminate the item list }

  Rst := $GETJPIW (Pidadr := Target_Pid,
		   Itmlst := Item_List,
		   Iosb   := Iosb) ;

  Zip (Rst,Iosb[1]) ;
  END ; 	{ of Get_Proc_Info }


[ASYNCHRONOUS] PROCEDURE Inject (Str : VARYING [Len] OF CHAR ;
				 Pseudo ,
				 See_This : BOOLEAN := TRUE) ;

{ Inject a string into a terminal buffer. }

  VAR  Choice , Rst : UNSIGNED := 0 ;

  BEGIN 	{ Inject }
  IF Bugger.D1	THEN
     BEGIN
     WRITELN (Dbgfile,'+++ INJECTING +++') ;
     WRITELN (Dbgfile,'PUSH: ',Str) ;
     END ;

  IF Pseudo  THEN  Choice := 1 ;

  Viewable := See_This ;

  CASE Choice  OF
    1: BEGIN
       FOR Rst := 1 TO Len  DO
	   Buffer.One[Wsts*2+2+Rst] := Str[Rst] ;

       IF Pseudo_Ft  THEN
	  BEGIN
	  Rst := PTD$WRITE (Chan       := Pchan,
			    Wrtbuf     := Buffer.Raw[Wsts],
			    Wrtbuf_Len := Len) ;
	  Zip (Rst) ;
	  END
	ELSE
	  BEGIN
{W}	  Rst := $QIOW (Chan := Pchan,
			Func := IO$_WRITEVBLK,
			P1   := Buffer.Raw[Wsts],
			P2   := Len) ;
	  Zip (Rst) ;
	  END ;
       END ;
    0: BEGIN
       IF Pseudo_Ft  THEN
	  BEGIN
	  FOR Rst := 1 TO Len  DO
	      Buffer.One[Rbuf*2+2+Rst] := Str[Rst] ;
	  Rst := $QIOW (Chan := Rchan,
			Func := IO$_WRITEVBLK,
			P1   := Buffer.Raw[Rbuf],
			P2   := Len) ;
	  Zip (Rst) ;
	  END
	ELSE
	  BEGIN
	  FOR Rst := 1 TO Len  DO
	      Buffer.One[Rsts*2+2+Rst] := Str[Rst] ;
	  Rst := $QIOW (Chan := Rchan,
			Func := IO$_WRITEVBLK,
			P1   := Buffer.Raw[Rsts],
			P2   := Len) ;
	 Zip (Rst) ;
	 END ;
       END ;
    END ;	{ of Case }

  IF Bugger.D1	THEN
     WRITELN (Dbgfile,'=== INJECTED ===') ;
  END ; 	{ Inject }


[ASYNCHRONOUS] PROCEDURE Rchan_Ast ;

{ Called when a keystroke occurs on the real keyboard - the keystrokes are
  passed to the pseudo-terminal and another read queued.  If escaping, avoid
  all read/write activity and simply wake up Mum. }

  VAR  Rst : UNSIGNED := 0 ;

  BEGIN 	{ Rchan_Ast}
  IF Escape_Hatch AND_THEN (Buffer.One[Wsts*2+3] = Ripcord)  THEN
     BEGIN
     LIB$SIGNAL (Jump__Userabort) ;		{ Bail out! }
     Jeronimo := TRUE ; 			{ Pull the ripcord! }
     Rst := $WAKE ;
     Zip (Rst) ;
     END
   ELSE
  IF Pseudo_Ft	THEN
     BEGIN
     Rst := PTD$WRITE (Chan	  := Pchan,
		       Wrtbuf	  := Buffer.Raw[Wsts],
		       Wrtbuf_Len := 1) ;
     Zip (Rst) ;

     Rst := $QIO (Chan	 := Rchan,
		  Func	 := IO$_READVBLK,
		  Astadr := Rchan_Ast,
		  P1	 := Buffer.Raw[Wbuf],
		  P2	 := 1) ;
     Zip (Rst) ;
     END
   ELSE
     BEGIN
     Rst := $QIO (Chan := Pchan,
		  Func := IO$_WRITEVBLK,
		  P1   := Buffer.Raw[Wsts],
		  P2   := 1) ;
     Zip (Rst) ;

     Rst := $QIO (Chan	 := Rchan,
		  Func	 := IO$_READVBLK,
		  Astadr := Rchan_Ast,
		  P1	 := Buffer.Raw[Wsts],
		  P2	 := 1) ;
     Zip (Rst) ;
     END ;
  END ; 	{ of Rchan_Ast }


[ASYNCHRONOUS] PROCEDURE Pchan_Ast ;

{ Called when characters are received from the pseudo-terminal - the data are
  passed to the real screen and another read queued. }

  CONST  Jump_Guillotine = 'JUMP_GUILLOTINE' ;

  VAR  Rst ,
       Null_Pos : UNSIGNED := 0 ;
       I , J , K , M : INTEGER := 0 ;
       Spawned ,
       Prompting : BOOLEAN := FALSE ;
       Current_Prompt : VARYING [255] OF CHAR := '' ;
       Log_Buffer : VARYING [SIZE(Io_Buffer)] OF CHAR := '' ;

       Was_Spawned   : [STATIC] BOOLEAN := FALSE ;
       Green_Process : [STATIC] BOOLEAN := TRUE ;
       Stage ,
       Cycle : [STATIC] INTEGER := 0 ;
       Dcl_Prompt : [STATIC] VARYING [255] OF CHAR := '' ;

  BEGIN 	{ Pchan_Ast }

  { Make it easy to work with the returned text. }

  IF Buffer.Raw[Rcnt] > 0  THEN
     STR$COPY_R (%DESCR Log_Buffer,Buffer.Raw[Rcnt],%REF Buffer.Raw[Rbuf])
   ELSE
     Log_Buffer := '' ;

  IF Single_Cmd_Mode  THEN
     BEGIN
     Cycle := Cycle + 1 ;
     IF Bugger.D1  THEN
	WRITELN (Dbgfile,'Pchan_Ast: --- Cycle ',Cycle:1,' ---') ;

     { Temporarily enable some extra privileges for the $GETJPI call. }

     $SETPRV (Enbflg := %IMMED 1,		{ Enable extra privs}
	      Prvadr := Single_Cmd_Mode_Privs) ;

     Get_Proc_Info (Pseudo_Pid,Proc_Cnt_Now,Job_Proc_Cnt_Now) ;

     $SETPRV (Enbflg := %IMMED 0,		{ Disable extra privs}
	      Prvadr := Single_Cmd_Mode_Privs) ;

     Spawned := (Proc_Cnt_Now > Proc_Cnt) OR
		(Job_Proc_Cnt_Now > Job_Proc_Cnt) ;
     Was_Spawned := Spawned OR Was_Spawned ;

     Null_Pos := INDEX (Log_Buffer,Null) ;

     { If debugging, write the new text to the debug file or
       dump it to the screen. }

     IF Bugger.D1  THEN
	BEGIN
	K := Log_Buffer.LENGTH DIV 20 ;
	M := Log_Buffer.LENGTH REM 20 ;
	WRITELN (Dbgfile,'BufLen: ',Log_Buffer.LENGTH:1) ;
	WRITELN (Dbgfile,'NULL Pos: ',Null_Pos:1) ;
	WRITE (Dbgfile,'Buffer:') ;
	FOR I := 1 TO K  DO
	    BEGIN
	    FOR J := 1 TO 20  DO
		WRITE (Dbgfile,HEX(Log_Buffer[(I-1)*20+J],3,2)) ;
	    WRITELN (Dbgfile) ;
	    IF M > 0  THEN  WRITE (Dbgfile,'       ') ;
	    END ;
	FOR I := 1 TO M  DO  WRITE (Dbgfile,HEX(Log_Buffer[K*20+I],3,2)) ;
	WRITELN (Dbgfile,Lf) ;
	END ;

     IF Bugger.D1 AND_THEN NOT Debug_To_Terminal  THEN
	  WRITELN (Dbgfile,'Buffer: ',Lf+Log_Buffer,Error:=CONTINUE) ;

     Prompting := ((Null_Pos = 2) AND_THEN
		    ((Log_Buffer[1] = Lf) OR (Log_Buffer[1] = Cr))) OR
		  ((Null_Pos = 3) AND_THEN
		    ((SUBSTR (Log_Buffer,1,2) = Cr+Lf) OR
			  (SUBSTR (Log_Buffer,1,2) = Lf+Cr))) ;
     IF Prompting  THEN
	Current_Prompt := SUBSTR (Log_Buffer,Null_Pos+1,
				  Log_Buffer.LENGTH-Null_Pos) ;

     { At process initiation, determine what the DCL prompt is. }

     IF Green_Process AND_THEN Prompting  THEN	  { Have not seen first prompt }
	BEGIN
	Dcl_Prompt := Current_Prompt ;
	Green_Process := FALSE ;
	IF Bugger.D1  THEN
	   WRITELN (Dbgfile,'Pchan_Ast: DCL_Prompt: ',Dcl_Prompt) ;
	END ;

     IF Prompting AND NOT Green_Process  THEN
	BEGIN
	Stage := Stage + 1 ;
	IF Bugger.D1  THEN
	   BEGIN
	   WRITELN (Dbgfile,'Prompt: ',Current_Prompt) ;
	   WRITELN (Dbgfile,'Stage:  ',Stage:1) ;
	   END ;

	CASE Stage OF
	  1: Inject ('SET PROMPT = '+Jump_Guillotine+Cr,,FALSE) ;
	  2: Inject (Lf+Cr+Null) ;
	  3: Inject ('$ '+Single_Cmd+Cr) ;
	  4: BEGIN
	     Inject (Cr) ;
	     IF Spawned  THEN
		LIB$SIGNAL (Jump__Nospawn) ;
	     END ;
	  5: Inject ('JUMP_SQUISHER :== EOJ'+Cr,,FALSE) ;
	  6: BEGIN
	     Inject ('JUMP_SQUISHER'+Cr,,FALSE) ;
	     IF Was_Spawned  THEN
		BEGIN
		Was_Spawned := FALSE ;
		Stage := 3 ;
		END ;
	     END ;
	  OTHERWISE LIB$SIGNAL (Jump__Intabort1) ; { Report to JUMP developer }
	  END ; { of Case }

	Visible := Viewable AND NOT (INDEX (Log_Buffer,Jump_Guillotine) > 0) ;
	END
      ELSE
	BEGIN
	Visible := Viewable ;
	Viewable := TRUE ;
	END ;

     IF Bugger.D1  THEN
	BEGIN
	WRITELN (Dbgfile,'Spawned:',Spawned) ;
	WRITELN (Dbgfile,'Visible:',Visible) ;
	IF Visible  THEN  WRITELN (Dbgfile) ;
	END ;
     END ;	{ Single_Cmd_Mode }

  { If recording, write the new text to the log file. }

  IF Record_Session AND Visible  THEN
     WRITELN (Logfile,Log_Buffer,Error:=CONTINUE) ;	{ Ignore errors }

  IF Pseudo_Ft	THEN
     BEGIN
     IF Visible  THEN
	BEGIN
	Rst := $QIOW (Chan := Rchan,
		      Func := IO$_WRITEVBLK,
		      P1   := Buffer.Raw[Rbuf],
		      P2   := Buffer.Raw[Rcnt]) ;
	Zip (Rst) ;
	END ;

     Rst := PTD$READ (Chan	  := Pchan,
		      Astadr	  := Pchan_Ast,
		      Readbuf	  := Buffer.Raw[Rsts],
		      Readbuf_Len := Ft_Buflen) ;
     Zip (Rst) ;
     END
   ELSE
     BEGIN
     IF Visible  THEN
	BEGIN
	Rst := $QIOW (Chan := Rchan,
		      Func := IO$_WRITEVBLK,
		      P1   := Buffer.Raw[Rsts],
		      P2   := Piosb[2]) ;
	Zip (Rst) ;
	END ;

     Rst := $QIO (Chan	 := Pchan,
		  Func	 := IO$_READVBLK,
		  Iosb	 := Piosb,
		  Astadr := Pchan_Ast,
		  P1	 := Buffer.Raw[Rsts],
		  P2	 := Py_Buflen) ;
     Zip (Rst) ;
     END ;
  END ; 	{ of Pchan_Ast }


[ASYNCHRONOUS] PROCEDURE Mbast ;

{ Invoked by the detached process termination mailbox AST - WAKE UP! }

  VAR  Rst : UNSIGNED := 0 ;

  BEGIN 	{ MBast }

  { Try to avoid any issues with typeahead buffer and EXACT log file }

  IF Record_Session  THEN
     BEGIN
     Rst := PTD$CANCEL (Pchan) ;
     Zip (Rst) ;
     END ;

  Rst := $WAKE ;
  Zip (Rst) ;
  END ; 	{ of MBast }


[ASYNCHRONOUS] PROCEDURE Send_Bell_Ast ;

{ Bell event notification AST. }

  VAR  Rst : UNSIGNED := 0 ;

  BEGIN 	{ Send_BELL_AST }
  Rst := $QIO (Chan := Rchan,
	       Func := IO$_WRITEVBLK,
	       P1   := Bell,
	       P2   := 1) ;
  Zip (Rst) ;
  END ; 	{ of Send_BELL_AST }


[ASYNCHRONOUS] PROCEDURE Send_Xon_Ast ;

{ Xon event notification AST. }

  VAR  Rst : UNSIGNED := 0 ;

  BEGIN 	{ Send_XON_AST }
  Rst := $QIO (Chan := Rchan,
	       Func := IO$_WRITEVBLK,
	       P1   := Xon,
	       P2   := 1) ;
  Zip (Rst) ;
  END ; 	{ of Send_XON_AST }


[ASYNCHRONOUS] PROCEDURE Send_Xoff_Ast ;

{ Xoff event notification AST. }

  VAR  Rst : UNSIGNED := 0 ;

  BEGIN 	{ Send_XOFF_AST }
  Rst := $QIO (Chan := Rchan,
	       Func := IO$_WRITEVBLK,
	       P1   := Xoff,
	       P2   := 1) ;
  Zip (Rst) ;
  END ; 	{ of Send_XOFF_AST }


PROCEDURE Transmography ;

{ Create a pseudo-terminal connected to a detached process and actually *be*
  the user!  This is JUMP /EXACT. }

  CONST

  Subversion_Msg = '**** JUMP WARNING!! Attempt to subvert Secure Mode ****' ;

  TYPE

  Desc_Blk = PACKED RECORD			{ Exit handler descriptor }
	       Fwd_Link : UNSIGNED ;
	       Exit_Handler_Addr : UNSIGNED ;
	       Argcnt : $UBYTE ;
	       Fill_Zero : [UNSAFE] ARRAY [1..3] OF $UBYTE ;
	       Condition ,
	       P2 , P3 , P4 , P5 , P6 , P7 : UNSIGNED ;
	     END VALUE ZERO ;

  VAR

  Modify_Sysprv : BOOLEAN := FALSE ;
  Rst , Mbunit , Stsflags : UNSIGNED := 0 ;
  Pctl : [VOLATILE,LONG] Prtctl_Type := ZERO ;
  Pctl_Ptr : [VOLATILE] Word_Ptr := NIL ;
  Rucb , Pucb ,
  Rapn, Papn: Unsigned_Ptr := NIL ;
  Newchars : Terminal_Chars := ZERO ;
  Exit_Desc : [STATIC] Desc_Blk := ZERO ;
  Iosb : Status_Block_Type := ZERO ;
  Item_List : Item_List_Template (2) := ZERO ;
  Specified_User : VARYING [Max_Username_Len] OF CHAR := '' ;
  Notify_Msg : VARYING [80] OF CHAR := '' ;

  FUNCTION Find_Device (Device : VARYING [L1] OF CHAR ;
			Chan   : $UWORD := 0) : BOOLEAN ;

  { Determine if a device exists and return its name. }

    VAR Rst : UNSIGNED := 0 ;
	Iosb : Status_Block_Type := ZERO ;
	Item_List : Item_List_Template (2) := ZERO ;

    BEGIN	{ Find_Device }
    Find_Device := TRUE ;

    Item_List[1].Buffer_Length := SIZE (Pdev.BODY) ;
    Item_List[1].Item_Code     := DVI$_DEVNAM ;
    Item_List[1].Buffer_Addr   := IADDRESS (Pdev.BODY) ;
    Item_List[1].Return_Addr   := IADDRESS (Pdev.LENGTH) ;

    Item_List[2].Terminator    := 0 ;		{ Terminate the item list }

    IF (Chan = 0) AND (Device <> '')  THEN	{ Device name supplied }
       Rst := $GETDVIW (Itmlst := Item_List,
			Devnam := Device,
			Iosb   := Iosb)
     ELSE					{ Channel number supplied }
       Rst := $GETDVIW (Itmlst := Item_List,
			Chan   := Chan,
			Iosb   := Iosb) ;

    IF Rst = SS$_NOSUCHDEV  THEN
       Find_Device := FALSE
     ELSE
    Zip (Rst,Iosb[1]) ;
    END ;	{ of Find_Device }

  PROCEDURE Get_Ucb (Chan_Num : UNSIGNED ;
		     VAR Ucb  : Unsigned_Ptr) ;

  { Use the logical UCB address in the CCB to get to the physical UCB.
    VAX and Alpha/Integrity handle things slightly differently. }

    VAR Inx : UNSIGNED := 0 ;
	Ccb_Chan : [VOLATILE] UNSIGNED := 0 ;
	Ccb_Chan_Ptr : Unsigned_Ptr := NIL ;

    BEGIN	{ Get_Ucb }
    CASE Architecture OF
      Vax:   BEGIN
	     Inx := Chan_Num DIV 16 * CCB$K_LENGTH ;
	     Ucb::UNSIGNED := CTL$GL_CCBBASE -
				Inx - CCB$L_UCB ;	{ Logical UCB address }
	     END ;
      Alpha,
      Integrity,
      X86: BEGIN
	     REPEAT
	       Ccb_Chan_Ptr::UNSIGNED := CTL$GA_CCB_TABLE + Inx + CCB$L_CHAN ;
	       Getmem (Ccb_Chan,Ccb_Chan_Ptr) ;
	       Inx := Inx + CCB$K_LENGTH ;
	     UNTIL Ccb_Chan = Chan_Num ;

	     Ucb::UNSIGNED := CTL$GA_CCB_TABLE + Inx - CCB$K_LENGTH +
			      CCB$L_UCB ;		{ Logical UCB address }
	     END ;
      END ;	{ of Case }

    Getmem (Ucb,Ucb) ;				{ Logical UCB }
    Ucb::UNSIGNED := Ucb::UNSIGNED +
		     UCB$L_TL_PHYUCB ;		{ Physical UCB address }
    Getmem (Ucb,Ucb) ;				{ Physical UCB }
    END ;	{ of Get_Ucb }

  BEGIN 	{ Transmography }
  IF Suspect  THEN			{ Secure Mode subversion attempt }
     BEGIN
     Oprmsg (Subversion_Msg,Oper_Class_Mask) ;
     Send_Mail_Message (Notify_Maillist,Subversion_Msg,'NL:') ;
     END ;

  Specified_User := New_User ;

  { Is external authentication being used? If so, use PERSONA routine.
    ***** ASSUMPTION *****
    The EXTAUTH flags will *only* be set if external authentication is
    either possibly or actually in use. }

  IF Psb_Available  THEN
     BEGIN
     Check_Extauth_Flags (Target_Flags) ;
     Use_Persona := Extauth_Flag ;
     Check_Extauth_Flags (Caller_Flags) ;
     Use_Persona := (Use_Persona OR Extauth_Flag) ;

     IF Bugger.Info  THEN
	WRITELN (Dbgfile,'*** Use_Persona:',Use_Persona) ;

     IF Use_Persona AND Bugger.X1  THEN
	BEGIN
	Use_Persona := FALSE ;
	LIB$SIGNAL (Jump__Notusepss) ;
	IF Bugger.Info	THEN
	   WRITELN (Dbgfile,'*** Use_Persona:',Use_Persona) ;
	END ;

     IF (NOT Use_Persona) AND Bugger.X2  THEN
	BEGIN
	Use_Persona := TRUE ;
	LIB$SIGNAL (Jump__Forcepss) ;
	IF Bugger.Info	THEN
	   WRITELN (Dbgfile,'*** Use_Persona:',Use_Persona) ;
	END ;
     END ;

  { Determine which pseudo-terminal type (if any) exists on the system. }

  Pseudo_Ft := Find_Device ('FTA0:') ;
  IF NOT Pseudo_Ft  THEN
     IF NOT Find_Device ('PYA0:')  THEN
	$EXIT (Jump__Nopseudo) ;

  { Get current (real) terminal process-specific device characteristics }

  Get_Channel ('TT:',Rchan) ;

  Rst := $QIOW (Chan := Rchan,
		Func := IO$_SENSEMODE,
		Iosb := Iosb,
		P1   := Rchars,
		P2   := 12) ;
  Zip (Rst,Iosb[1]) ;

  { Audit close to, but before the guts of it, so that even if the jump fails,
    the attempt is audited. }

  IF Auditing  THEN
     Audit_Jump (TRUE) ;

  { If recording, construct the session log filename and make sure we can
    open the file early on. }

  IF Record_Session THEN
     BEGIN
     GETTIMESTAMP (Time_Now) ;
     Time_Str := DATE (Time_Now) + ' ' + TIME (Time_Now) ;
     IF Time_Str[1]  = ' '  THEN
	Time_Str[1] := '0' ;

     IF NOT Secure_Mode  THEN
	Secure_Directory := User_Directory ;
     WRITEV (Session_Log,Secure_Directory,
			 'JUMP_',
			 Orig_User,
			 '-',
			 New_User,
			 '.',
			 DEC (Time_Now.Year,4,4),
			 DEC (Time_Now.Month,2,2),
			 DEC (Time_Now.Day,2,2),
			 '_',
			 DEC (Time_Now.Hour,2,2),
			 DEC (Time_Now.Minute,2,2),
			 DEC (Time_Now.Second,2,2) ) ;

     { If the user has specified JUMP_USER_DIR and the System Manager has
       not, turn off SYSPRV for the file open if SYSPRV is not in the user's
       process privilege mask. }

     Modify_Sysprv := NOT (Secure_User_Dir OR (Sysprv IN Proc_Perm_Priv)) ;

     IF Modify_Sysprv  THEN
	$SETPRV (Enbflg := %IMMED 0,		{ Disable SYSPRV }
		 Prvadr := PRV$M_SYSPRV) ;

     Open_New_File := TRUE ;
     OPEN (Logfile,FILE_NAME:=Session_Log,HISTORY:=NEW,
		   Record_Length:=SIZE(Io_Buffer),Carriage_Control:=None,
		   USER_ACTION:=Special_File_Open,Error:=CONTINUE) ;
     Rst := STATUS (Logfile) ;
     IF Rst <> PAS$K_SUCCESS  THEN
	BEGIN
	LIB$SIGNAL (Jump__Badlogfil) ;
	LIB$STOP (Condition_Code (Rst)) ;
	END ;

     REWRITE (Logfile) ;
     Session_Log := Open_File_Name ;

     { If necessary, restore SYSPRV. }

     IF Modify_Sysprv  THEN
	$SETPRV (Enbflg := %IMMED 1,		{ Enable SYSPRV }
		 Prvadr := PRV$M_SYSPRV) ;
     END ;

  { Set up and declare the exit handler }

  Exit_Desc.Exit_Handler_Addr := IADDRESS (Exit_Handler) ;
  Exit_Desc.Argcnt := 1 ;
  Exit_Desc.Condition := IADDRESS (Exit_Rst) ;

  Rst := $DCLEXH (Desblk := Exit_Desc) ;
  Zip (Rst) ;

  { If required, issue the notification. }

  IF Notify.Before  THEN
     BEGIN
     Notify_Msg := 'Initiated JUMP/EXACT to ' + New_User ;
     IF Notify.By_Opcom  THEN
	Oprmsg (Notify_Msg,Oper_Class_Mask) ;
     IF Notify.By_Mail	THEN
	Send_Mail_Message (Notify_Maillist,Notify_Msg,'NL:') ;
     END ;

  { Set new terminal characteristics }

  Newchars := Rchars ;
  Newchars.Tt_Devchar.TT$V_NOECHO    := TRUE ;
  Newchars.Tt_Devchar.TT$V_WRAP      := FALSE ;
  Newchars.Tt_Devchar2.TT2$V_PASTHRU := TRUE ;

  Rst := $QIOW (Chan := Rchan,
		Func := IO$_SETMODE,
		Iosb := Iosb,
		P1   := Newchars,
		P2   := 12) ;
  Zip (Rst,Iosb[1]) ;

  IF Use_Persona  THEN
     Change_Persona (New_User)
   ELSE
     BEGIN
     Change_Username (New_User) ;	{ *** Change to new username *** }
     Change_Uic (New_Uic) ;		{ *** Change to new UIC *** }
     END ;

  IF Bugger.X9	THEN  $EXIT (Jump__Abortx9) ;{ Bug out before we get serious }

  IF Bugger.D4	THEN
     WRITELN (Dbgfile,'*** Create Pseudoterm ...') ;

  { Create the pseudo-terminal }

  Pbuf_Range[1] := IADDRESS (Buffer) ;
  Pbuf_Range[2] := Pbuf_Range[1] + Io_Buflen * 2 - 1 ;

  IF Pseudo_Ft	THEN
     BEGIN
     Rst := PTD$CREATE (Chan	 := Pchan,
			Charbuff := Rchars,
			Buflen	 := SIZE (Rchars),
			Inadr	 := Pbuf_Range) ;
     Zip (Rst) ;

     { Set event notifications }

     Rst := PTD$SET_EVENT_NOTIFICATION (Chan   := Pchan,
					Astadr := Send_Bell_Ast,
					Type_  := PTD$C_SEND_BELL) ;
     Zip (Rst) ;

     Rst := PTD$SET_EVENT_NOTIFICATION (Chan   := Pchan,
					Astadr := Send_Xon_Ast,
					Type_  := PTD$C_SEND_XON) ;
     Zip (Rst) ;

     Rst := PTD$SET_EVENT_NOTIFICATION (Chan   := Pchan,
					Astadr := Send_Xoff_Ast,
					Type_  := PTD$C_SEND_XOFF) ;
     Zip (Rst) ;
     END
   ELSE
     BEGIN
     Get_Channel ('PYA0:',Pchan) ;
     END ;

  Find_Device ('',Pchan) ;		{ Sets Pdev to device name }
  Pchan_Created := TRUE ;

  { If the "real" process has a valid value for ACCPORNAM, set the "pseudo"
    process to point to it. }

  Get_Ucb (Pchan,Pucb) ;
  Pctl_Ptr::UNSIGNED := Pucb::UNSIGNED + UCB$W_TT_PRTCTL ;
  Getmem (Pctl,Pctl_Ptr) ;
  Papn::UNSIGNED := Pucb::UNSIGNED + UCB$L_TT_ACCPORNAM ;

  IF Port = ''	THEN
     BEGIN

     { Welcome to a futureware section of code!  When ACCPORNAM is not valid,
       the port name will be empty.  In this case, use the terminal name.
       Allocate an appropriate buffer, copy the terminal name into it,
       put the address of the buffer into ACCPORNAM of the pseudo-terminal
       and set the validity bit in the port control mask.  When the pseudo-
       terminal is terminated, deallocate the buffer.  This exercise remains to
       be attempted ... or supplied by an eager code jockey somewhere else! :) }

     END
   ELSE
     BEGIN
     Get_Ucb (Rchan,Rucb) ;
     Rapn::UNSIGNED := Rucb::UNSIGNED + UCB$L_TT_ACCPORNAM ;
     Getmem (Rapn,Rapn) ;
     Putmem (Rapn,Papn) ;
     Pctl::Prtctl_Type.TTY$V_PC_ACCPORNAM := TRUE ;
     Putmem (Pctl,Pctl_Ptr) ;
     END ;

  IF Bugger.D4	THEN
     WRITELN (Dbgfile,'*** Create Termination MBX ...') ;

  { Create a termination mailbox for the soon-to-be detached process,
    and get its unit number }

  Rst := $CREMBX (Chan	 := Mchan,
		  Maxmsg := ACC$K_TERMLEN) ;
  Zip (Rst) ;

  Item_List[1].Buffer_Length := 4 ;
  Item_List[1].Item_Code     := DVI$_UNIT ;
  Item_List[1].Buffer_Addr   := IADDRESS (Mbunit) ;
  Item_List[1].Return_Addr   := 0 ;

  Item_List[2].Terminator    := 0 ;	{ Terminate the item list }

  Rst := $GETDVIW (Itmlst := Item_List,
		   Chan   := Mchan,
		   Iosb   := Iosb) ;
  Zip (Rst,Iosb[1]) ;

  { Queue an asynchronous read to the mailbox }

  Rst := $QIO (Chan   := Mchan,
	       Func   := IO$_READVBLK,
	       Astadr := Mbast,
	       P1     := Mbbuf,
	       P2     := ACC$K_TERMLEN) ;
  Zip (Rst) ;

  IF Bugger.D4	THEN
     WRITELN (Dbgfile,'*** Create Detached Process ...') ;

  { Create the detached process }

  Stsflags := PRC$M_DETACH + PRC$M_INTER + PRC$M_NOPASSWORD ;
  IF Figment  THEN
     Stsflags := Stsflags + PRC$M_LOGIN ;	{ == PRC$M_NOUAF }

  Rst := $CREPRC (Pidadr := %REF Pseudo_Pid,
		  Image  := 'SYS$SYSTEM:LOGINOUT.EXE',
		  INPUT  := Pdev,
		  OUTPUT := Pdev,
		  Error  := Pdev,
		  Baspri := 4,
		  Mbxunt := Mbunit,
		  Prcnam := 'JUMP_' + HEX(Pid,8,8),
		  Stsflg := %IMMED Stsflags) ;

  Zip (Rst) ;

  IF Stamp_Uaf	THEN  Update_Uaf ;	{ Update UAF login details }

  IF NOT Chain	THEN			{ Chaining not allowed }
     Target_Attributes := Target_Attributes + 'NOCHAIN ' ;

  IF Target_Attributes = ''  THEN
     Target_Attributes := 'NONE'
   ELSE
     Define_Logical_Name ('JUMP_'+HEX(Pseudo_Pid,8,8),Target_Attributes) ;

  IF Bugger.Info OR Bugger.D4  THEN
     BEGIN
     WRITELN (Dbgfile,'*** Target PID: ',HEX(Pseudo_Pid,8,8)) ;
     WRITELN (Dbgfile,'*** Target Attributes: ',Target_Attributes) ;
     WRITELN (Dbgfile) ;
     END ;

  IF Log  THEN
     BEGIN
     LIB$SIGNAL (Jump__Transfer,1,
		 %STDESCR SUBSTR (New_User,1,New_User.LENGTH)) ;
     WRITELN ;
     IF Single_Cmd_Mode  THEN  WRITELN ;  { Again }
     END ;

  { Restore username and UIC }

  New_User.BODY := PAD ('',' ',Max_Username_Len) ;	{ Totally blat it }
  New_User := Orig_User ;
  New_Uic  := Orig_Uic ;

  IF Use_Persona  THEN
     Change_Persona (New_User)
   ELSE
     BEGIN
     Change_Username (New_User) ;	{ *** Change to new username *** }
     Change_Uic (New_Uic) ;		{ *** Change to new UIC *** }
     END ;

  { Put some audit information at the start of the session log file. Note
    that we have to put the carriage control in explicitly because the
    log file is created without any implied carriage control. }

  IF Record_Session THEN
     BEGIN
     WRITELN (Logfile,PAD ('-','-',78),Cr) ;
     WRITELN (Logfile,Lf,Jump_Version,Cr) ;
     WRITELN (Logfile,Lf,'Pseudo-terminal session log.',Cr) ;
     WRITELN (Logfile,Lf,Session_Log,Cr) ;
     WRITELN (Logfile,Lf,'User:        ',Orig_User,Cr) ;
     WRITELN (Logfile,Lf,'Login time:  ', Login_Time_Str, Cr) ;
     WRITELN (Logfile,Lf,'From PID:    ',HEX (Master_Pid,8,8),Cr) ;
     WRITELN (Logfile,Lf,'Process:     ',Process_Name,Cr) ;
     WRITELN (Logfile,Lf,'To PID:      ',HEX (Pseudo_Pid,8,8),Cr) ;
     IF Physical_Device <> ''  THEN
	WRITELN (Logfile,Lf,'Phys Dev:    ',Physical_Device,Cr) ;
     IF Terminal <> ''	THEN
	WRITELN (Logfile,Lf,'Terminal:    ',Terminal,Cr) ;
     IF Port <> ''  THEN
	WRITELN (Logfile,Lf,'Port:        ',Port,Cr) ;
     WRITELN (Logfile,Lf,'JUMP time:   ',Time_Str,Cr) ;
     WRITELN (Logfile,Lf,'Target user: ',Specified_User,Cr) ;
     IF Secure_Mode  THEN
	WRITELN (Logfile,Lf,'Mode:        Secure',Cr)
      ELSE
	WRITELN (Logfile,Lf,'Mode:        Record',Cr) ;
     WRITELN (Logfile,Lf,'Attributes:  ',Target_Attributes,Cr) ;
     WRITELN (Logfile,Lf,'Command:     ',Command,Cr) ;
     IF Suspect  THEN
	WRITELN (Logfile,Lf,Subversion_Msg,Cr) ;
     WRITELN (Logfile,PAD (Lf,'-',78),Cr) ;
     WRITELN (Logfile,Lf,Cr) ;
     END ;

  { Queue the appropriate reads to both the real terminal and pseudo-terminal }

  IF Pseudo_Ft	THEN
     BEGIN
     Rst := $QIO (Chan	 := Rchan,
		  Func	 := IO$_READVBLK,
		  Astadr := Rchan_Ast,
		  P1	 := Buffer.Raw[Wbuf],
		  P2	 := 1) ;
     Zip (Rst) ;

     Rst := PTD$READ (Chan	  := Pchan,
		      Astadr	  := Pchan_Ast,
		      Readbuf	  := Buffer.Raw[Rsts],
		      Readbuf_Len := Ft_Buflen) ;
     Zip (Rst) ;
     END
   ELSE
     BEGIN
     Rst := $QIO (Chan	 := Rchan,
		  Func	 := IO$_READVBLK,
		  Astadr := Rchan_Ast,
		  P1	 := Buffer.Raw[Wsts],
		  P2	 := 1) ;
     Zip (Rst) ;

     Rst := $QIO (Chan	 := Pchan,
		  Func	 := IO$_READVBLK,
		  Iosb	 := Piosb,
		  Astadr := Pchan_Ast,
		  P1	 := Buffer.Raw[Rsts],
		  P2	 := Py_Buflen) ;
     Zip (Rst) ;
     END ;

  { Hibernate until termination mailbox message wakes us up. }

  $HIBER ;

  { All Done! We're back! ...

    If recording session for an EXACT jump, close the log file. If the escape
    character terminated the session, add an appropriate termination message
    to the session log before closing it. }

  IF Record_Session  THEN
     BEGIN
     IF Jeronimo  THEN
	WRITELN (Logfile,Lf,'************ ',
			    'Process terminated by user escape request ',
			    '************',Cr,Error:=CONTINUE) ;
     CLOSE (Logfile,Error:=CONTINUE) ;
     END ;

  { Same deal for debug as for recording }

  IF Debug AND NOT Debug_To_Terminal  THEN      { Not if it's a terminal }
     BEGIN
     IF Jeronimo  THEN
	BEGIN
	WRITELN (Dbgfile) ;
	WRITELN (Dbgfile,'************ ',
			 'Process terminated by user escape request ',
			 '************',Error:=CONTINUE) ;
	END ;
     CLOSE (Dbgfile,Error:=CONTINUE) ;
     END ;

  { If required, notify the troops. }

  IF Notify.After  THEN
     BEGIN
     Notify_Msg := 'Completed JUMP/EXACT to ' + Specified_User ;
     IF Notify.By_Mail	THEN
	BEGIN
	IF Notify.Include_Log  THEN
	   Send_Mail_Message (Notify_Maillist,Notify_Msg,Session_Log)
	 ELSE
	IF Record_Session  THEN
	   Send_Mail_Message (Notify_Maillist,Notify_Msg,
			      'The JUMP session log is ' + Session_Log)
	 ELSE
	   Send_Mail_Message (Notify_Maillist,Notify_Msg,
			      'There was no JUMP session log.') ;
	END ;
     IF Notify.By_Opcom  THEN
	BEGIN
	Oprmsg (Notify_Msg,Oper_Class_Mask) ;
	IF Record_Session  THEN
	   Oprmsg ('JUMP Session log: ' + Session_Log,Oper_Class_Mask) ;
	END ;
     END ;

  WRITELN ;
  IF Log  THEN
     BEGIN
     LIB$SIGNAL (Jump__Return,1,
		 %STDESCR SUBSTR (Orig_User,1,Orig_User.LENGTH)) ;
     WRITELN ;
     END ;

  { The Exit_Handler will do the cleanup. }

  END ; 	{ of Transmography }


{ * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
  * * * * * * * * * *	M A I N   P R O G R A M   * * * * * * * * * *
  * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * }


BEGIN	{ Jump }

Get_System_Info ;

Get_Caller_Info ;

Get_And_Parse_Command ;

IF Bugger.X3  THEN
   ESTABLISH (General_Error_Handler) ;

IF Bugger.X6  THEN
   BEGIN
   LIB$SIGNAL (Jump__Crashrequested) ;
   Appcrash := Appcrash/Appcrash ; { Deliberately crash JUMP }
   END ;

Validate_Access ;			{ Stop intruders! }

IF NOT Production_Version  THEN  LIB$SIGNAL (Jump__Devnotprd) ;

IF Real_Mccoy  THEN
   Transmography			{ Clone! }
 ELSE
   BEGIN
   IF Auditing	THEN
      Audit_Jump (TRUE)
    ELSE
      Format_User (New_Uic) ;

   IF Alter_Ego  THEN			{ *** Change username *** }
      Change_Username (New_User) ;

   IF Transmute  THEN			{ Long jump! Boing! }
      Kangaroo ;			{ *** Change miscellany of items *** }

   IF Log  THEN  Display_Jump ;
   END ;

END.	{ of it all }
